/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var Histogram1445664487616;
        (function (Histogram1445664487616) {
            // jsCommon
            var PixelConverter = jsCommon.PixelConverter;
            var createClassAndSelector = jsCommon.CssConstants.createClassAndSelector;
            var VisualDataRoleKind = powerbi.VisualDataRoleKind;
            var ValueType = powerbi.ValueType;
            var DateTimeSequence = powerbi.DateTimeSequence;
            var DataViewObjects = powerbi.DataViewObjects;
            var TextMeasurementService = powerbi.TextMeasurementService;
            // powerbi.visuals
            var ValueFormatter = powerbi.visuals.valueFormatter;
            var SelectionId = powerbi.visuals.SelectionId;
            var axisStyle = powerbi.visuals.axisStyle;
            var yAxisPosition = powerbi.visuals.yAxisPosition;
            var DataColorPalette = powerbi.visuals.DataColorPalette;
            var ColorHelper = powerbi.visuals.ColorHelper;
            var SVGUtil = powerbi.visuals.SVGUtil;
            var TooltipManager = powerbi.visuals.TooltipManager;
            var dataLabelUtils = powerbi.visuals.dataLabelUtils;
            var willLabelsFit = powerbi.visuals.AxisHelper.LabelLayoutStrategy.willLabelsFit;
            var willLabelsWordBreak = powerbi.visuals.AxisHelper.LabelLayoutStrategy.willLabelsWordBreak;
            var axisScale = powerbi.visuals.axisScale;
            var valueFormatter = powerbi.visuals.valueFormatter;
            var ObjectEnumerationBuilder = powerbi.visuals.ObjectEnumerationBuilder;
            var appendClearCatcher = powerbi.visuals.appendClearCatcher;
            var createInteractivityService = powerbi.visuals.createInteractivityService;
            var HistogramChartWarning = (function () {
                function HistogramChartWarning(message) {
                    this.message = message;
                }
                Object.defineProperty(HistogramChartWarning.prototype, "code", {
                    get: function () {
                        return "HistogramChartWarning";
                    },
                    enumerable: true,
                    configurable: true
                });
                HistogramChartWarning.prototype.getMessages = function (resourceProvider) {
                    return {
                        message: this.message,
                        title: resourceProvider.get(""),
                        detail: resourceProvider.get("")
                    };
                };
                HistogramChartWarning.ErrorInvalidDataValues = "Some data values are invalid or too big";
                return HistogramChartWarning;
            }());
            Histogram1445664487616.HistogramChartWarning = HistogramChartWarning;
            var Histogram = (function () {
                function Histogram(constructorOptions) {
                    this.widthOfColumn = 0;
                    this.yTitleMargin = 0;
                    this.outerPadding = 0;
                    if (constructorOptions) {
                        this.svg = constructorOptions.svg;
                        this.animator = constructorOptions.animator;
                        this.margin = constructorOptions.margin;
                        this.behavior = constructorOptions.behavior;
                    }
                    this.svg = this.svg || undefined;
                    this.animator = this.animator || undefined;
                    this.margin = this.margin || Histogram.DefaultMargin;
                    this.behavior = this.behavior || HistogramBehavior.create();
                }
                Object.defineProperty(Histogram.prototype, "columnsSelection", {
                    get: function () {
                        return this.main.select(Histogram.Columns.selector)
                            .selectAll(Histogram.Column.selector);
                    },
                    enumerable: true,
                    configurable: true
                });
                Histogram.prototype.init = function (visualsOptions) {
                    this.hostService = visualsOptions.host;
                    this.interactivityService = createInteractivityService(this.hostService);
                    if (this.svg) {
                        this.root = this.svg;
                    }
                    else {
                        this.root = d3.select(visualsOptions.element.get(0))
                            .append("svg");
                    }
                    var style = visualsOptions.style;
                    this.colors = style && style.colorPalette
                        ? style.colorPalette.dataColors
                        : new DataColorPalette();
                    this.clearCatcher = appendClearCatcher(this.root);
                    this.root.classed(Histogram.ClassName, true);
                    this.main = this.root.append("g");
                    this.columns = this.main
                        .append("g")
                        .classed(Histogram.Columns.class, true);
                    this.axes = this.main
                        .append("g")
                        .classed(Histogram.Axes.class, true);
                    this.axisX = this.axes
                        .append("g")
                        .classed(Histogram.Axis.class, true)
                        .classed(Histogram.XAxis.class, true);
                    this.axisY = this.axes
                        .append("g")
                        .classed(Histogram.Axis.class, true)
                        .classed(Histogram.YAxis.class, true);
                    this.legend = this.main
                        .append("g")
                        .classed(Histogram.Legends.class, true);
                    this.labelGraphicsContext = this.main
                        .append("g")
                        .classed(Histogram.LabelGraphicsContext.class, true);
                };
                Histogram.converter = function (dataView, colors) {
                    if (!dataView ||
                        !dataView.categorical ||
                        !dataView.categorical.categories ||
                        !dataView.categorical.categories[0] ||
                        !dataView.categorical.categories[0].values ||
                        !(dataView.categorical.categories[0].values.length > 0) ||
                        !colors) {
                        return null;
                    }
                    var settings, histogramLayout, values, numericalValues = [], bins, dataPoints, valueFormatter, frequencies = [], identities = [], shiftByValues = 0, sumFrequency = 0, xLabelFormatter, yLabelFormatter, xLegendSize, yLegendSize, borderValues, yAxisSettings, sourceValues = dataView.categorical.categories[0].values;
                    settings = Histogram.parseSettings(dataView, colors);
                    if (!settings
                        || !Histogram.areValuesNumbers(dataView.categorical.categories[0])
                        || sourceValues.length < Histogram.MinAmountOfValues) {
                        return null;
                    }
                    xLegendSize = Histogram.getLegendSize(settings.xAxisSettings);
                    yLegendSize = Histogram.getLegendSize(settings.yAxisSettings);
                    if (dataView.categorical.values &&
                        dataView.categorical.values[0] &&
                        dataView.categorical.values[0].values) {
                        frequencies = dataView.categorical.values[0].values;
                    }
                    if (dataView.categorical.categories[0].identity
                        && dataView.categorical.categories[0].identity.length > 0) {
                        identities = dataView.categorical.categories[0].identity;
                    }
                    values = Histogram.getValuesByFrequencies(sourceValues, frequencies, identities);
                    values.forEach(function (value) {
                        numericalValues.push(value.value);
                        sumFrequency += value.frequency;
                    });
                    histogramLayout = d3.layout.histogram();
                    if (settings.bins && settings.bins > Histogram.MinNumberOfBins) {
                        histogramLayout = histogramLayout.bins(settings.bins);
                    }
                    bins = histogramLayout.frequency(settings.frequency)(numericalValues);
                    bins.forEach(function (bin, index) {
                        var filteredValues, frequency;
                        filteredValues = values.filter(function (value) {
                            return Histogram.isValueContainedInRange(value, bin, index);
                        });
                        frequency = filteredValues.reduce(function (previousValue, currentValue) {
                            return previousValue + currentValue.frequency;
                        }, 0);
                        bin.y = settings.frequency
                            ? frequency
                            : frequency / sumFrequency;
                        shiftByValues += bin.length;
                    });
                    borderValues = Histogram.getBorderValues(bins);
                    yAxisSettings = settings.yAxisSettings;
                    var maxYvalue = (yAxisSettings.end !== null) && (yAxisSettings.end > yAxisSettings.start)
                        ? yAxisSettings.end
                        : borderValues.maxY;
                    var minYValue = yAxisSettings.start < maxYvalue
                        ? yAxisSettings.start
                        : 0;
                    settings.yAxisSettings.start = Histogram.getCorrectXAxisValue(minYValue);
                    settings.yAxisSettings.end = Histogram.getCorrectXAxisValue(maxYvalue);
                    if (values.length >= Histogram.MinAmountOfValues) {
                        valueFormatter = ValueFormatter.create({
                            format: ValueFormatter.getFormatString(dataView.categorical.categories[0].source, Histogram.Properties["general"]["formatString"]),
                            value: values[0].value,
                            value2: values[values.length - 1].value,
                            precision: settings.precision
                        });
                        xLabelFormatter = ValueFormatter.create({
                            value: settings.xAxisSettings.displayUnits === 0
                                ? values[values.length - 1].value
                                : settings.xAxisSettings.displayUnits,
                            precision: settings.xAxisSettings.precision
                        });
                        yLabelFormatter = ValueFormatter.create({
                            value: settings.yAxisSettings.displayUnits,
                            precision: settings.yAxisSettings.precision
                        });
                    }
                    dataPoints = Histogram.getDataPoints(values, numericalValues, bins, settings, yLabelFormatter, xLabelFormatter);
                    return {
                        dataPoints: dataPoints,
                        borderValues: borderValues,
                        settings: settings,
                        formatter: valueFormatter,
                        xLabelFormatter: xLabelFormatter,
                        yLabelFormatter: yLabelFormatter,
                        xLegendSize: xLegendSize,
                        yLegendSize: yLegendSize
                    };
                };
                Histogram.getBorderValues = function (bins) {
                    var borderValues = {
                        minX: Number.MAX_VALUE,
                        maxX: -Number.MAX_VALUE,
                        minY: Number.MAX_VALUE,
                        maxY: -Number.MAX_VALUE
                    };
                    bins.forEach(function (dataPoint) {
                        var minX = Number.MAX_VALUE, maxX = -Number.MAX_VALUE;
                        dataPoint.forEach(function (x) {
                            if (x > maxX) {
                                maxX = x;
                            }
                            if (x < minX) {
                                minX = x;
                            }
                        });
                        if (minX < borderValues.minX) {
                            borderValues.minX = minX;
                        }
                        if (maxX > borderValues.maxX) {
                            borderValues.maxX = maxX;
                        }
                        if (dataPoint.y < borderValues.minY) {
                            borderValues.minY = dataPoint.y;
                        }
                        if (dataPoint.y > borderValues.maxY) {
                            borderValues.maxY = dataPoint.y;
                        }
                    });
                    return borderValues;
                };
                Histogram.getCorrectXAxisValue = function (value) {
                    return Math.max(Math.min(value, Histogram.MaxXAxisEndValue), Histogram.MinXAxisStartValue);
                };
                Histogram.areValuesNumbers = function (categoryColumn) {
                    return categoryColumn
                        && categoryColumn.source
                        && (categoryColumn.source.type.numeric || categoryColumn.source.type.integer);
                };
                Histogram.getValuesByFrequencies = function (sourceValues, frequencies, identities) {
                    var values = [];
                    sourceValues.forEach(function (item, index) {
                        var frequency = 1, value = Number(item);
                        value = isNaN(value) ? 0 : value;
                        if (frequencies
                            && frequencies[index]
                            && !isNaN(frequencies[index])
                            && frequencies[index] > 1) {
                            frequency = frequencies[index];
                        }
                        values.push({
                            value: value,
                            frequency: frequency,
                            selectionId: SelectionId.createWithId(identities[index])
                        });
                    });
                    return values;
                };
                Histogram.getDataPoints = function (values, numericalValues, bins, settings, yValueFormatter, xValueFormatter) {
                    var fontSizeInPx = PixelConverter.fromPoint(settings.labelSettings.fontSize);
                    return bins.map(function (bin, index) {
                        bin.range = Histogram.getRange(bin.x, bin.dx);
                        bin.tooltipInfo = Histogram.getTooltipData(bin.y, bin.range, settings, index === 0, yValueFormatter, xValueFormatter);
                        bin.subDataPoints = Histogram.getSubDataPoints(values, bin, index);
                        bin.labelFontSize = fontSizeInPx;
                        return bin;
                    });
                };
                Histogram.getRange = function (x, dx) {
                    return [x, x + dx];
                };
                Histogram.getTooltipData = function (value, range, settings, includeLeftBorder, yValueFormatter, xValueFormatter) {
                    return [
                        {
                            displayName: Histogram.getLegendText(settings),
                            value: yValueFormatter.format(value)
                        }, {
                            displayName: Histogram.TooltipDisplayName,
                            value: Histogram.rangeToString(range, includeLeftBorder, xValueFormatter)
                        }
                    ];
                };
                Histogram.getSubDataPoints = function (values, bin, index) {
                    var dataPoints = [];
                    values.forEach(function (value) {
                        if (Histogram.isValueContainedInRange(value, bin, index)) {
                            dataPoints.push({
                                identity: value.selectionId,
                                selected: false
                            });
                        }
                    });
                    return dataPoints;
                };
                Histogram.isValueContainedInRange = function (value, bin, index) {
                    return ((index === 0 && value.value >= bin.x) || (value.value > bin.x)) && value.value <= bin.x + bin.dx;
                };
                Histogram.parseSettings = function (dataView, colors) {
                    if (!dataView ||
                        !dataView.metadata ||
                        !dataView.metadata.columns ||
                        !dataView.metadata.columns[0]) {
                        return null;
                    }
                    var histogramSettings = {}, objects, colorHelper;
                    colorHelper = new ColorHelper(colors, Histogram.Properties["dataPoint"]["fill"], Histogram.DefaultHistogramSettings.fillColor);
                    histogramSettings.displayName =
                        dataView.metadata.columns[0].displayName || Histogram.DefaultHistogramSettings.displayName;
                    objects = Histogram.getObjectsFromDataView(dataView);
                    var xAxisSettings = {
                        axisColor: Histogram.getXAxisColor(objects).solid.color,
                        title: Histogram.getXTitle(objects),
                        precision: Histogram.getXPrecision(objects),
                        style: Histogram.getXStyle(objects),
                        displayUnits: Histogram.getXDisplayUnit(objects),
                        show: Histogram.getXAxisShow(objects),
                    };
                    var yAxisSettings = {
                        axisColor: Histogram.getYAxisColor(objects).solid.color,
                        title: Histogram.getYTitle(objects),
                        precision: Histogram.getYPrecision(objects),
                        style: Histogram.getYStyle(objects),
                        displayUnits: Histogram.getYDisplayUnit(objects),
                        show: Histogram.getYAxisShow(objects),
                        start: Histogram.getYStart(objects),
                        end: Histogram.getYEnd(objects),
                        position: Histogram.getYPosition(objects),
                    };
                    var labelSettings = {
                        show: Histogram.getLabelShow(objects),
                        color: Histogram.getLabelColor(objects).solid.color,
                        displayUnits: Histogram.getLabelDisplayUnits(objects),
                        precision: Histogram.getLabelPrecision(objects),
                        fontSize: Histogram.getLabelFontSize(objects),
                    };
                    histogramSettings.fillColor = colorHelper.getColorForMeasure(objects, "");
                    histogramSettings.bins = Histogram.getBins(objects);
                    histogramSettings.frequency = Histogram.getFrequency(objects);
                    histogramSettings.precision = Histogram.getPrecision(objects);
                    histogramSettings.displayName = Histogram.getLegend(histogramSettings.displayName, xAxisSettings.style, xAxisSettings.displayUnits);
                    histogramSettings.xAxisSettings = xAxisSettings;
                    histogramSettings.yAxisSettings = yAxisSettings;
                    histogramSettings.labelSettings = labelSettings;
                    return histogramSettings;
                };
                Histogram.getLegend = function (title, style, displayUnit) {
                    var retValue, formatter;
                    formatter = ValueFormatter.create({
                        value: displayUnit
                    });
                    switch (style) {
                        case axisStyle.showTitleOnly: {
                            retValue = title;
                            break;
                        }
                        case axisStyle.showUnitOnly: {
                            retValue = !(displayUnit === 0 || displayUnit === 1) && formatter.displayUnit
                                ? formatter.displayUnit.title
                                : title;
                            break;
                        }
                        case axisStyle.showBoth: {
                            retValue = !(displayUnit === 0 || displayUnit === 1) && formatter.displayUnit
                                ? title + " (" + formatter.displayUnit.title + ")"
                                : title;
                            break;
                        }
                    }
                    return retValue;
                };
                Histogram.getLabelFontSize = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["labels"]["fontSize"], Histogram.DefaultHistogramSettings.labelSettings.fontSize);
                };
                Histogram.getLabelShow = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["labels"]["show"], Histogram.DefaultHistogramSettings.labelSettings.show);
                };
                Histogram.getLabelColor = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["labels"]["color"], {
                        solid: {
                            color: Histogram.DefaultHistogramSettings.labelSettings.color
                        }
                    });
                };
                Histogram.getLabelDisplayUnits = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["labels"]["displayUnits"], Histogram.DefaultHistogramSettings.labelSettings.displayUnits);
                };
                Histogram.getLabelPrecision = function (objects) {
                    var precision = DataViewObjects.getValue(objects, Histogram.Properties["labels"]["precision"], Histogram.DefaultHistogramSettings.labelSettings.precision);
                    if (precision <= Histogram.MinPrecision) {
                        return Histogram.MinPrecision;
                    }
                    else if (precision >= Histogram.MaxPrecision) {
                        return Histogram.MaxPrecision;
                    }
                    return precision;
                };
                Histogram.getXStyle = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["style"], Histogram.DefaultHistogramSettings.xAxisSettings.style);
                };
                Histogram.getXDisplayUnit = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["displayUnits"], Histogram.DefaultHistogramSettings.xAxisSettings.displayUnits);
                };
                Histogram.getXPrecision = function (objects) {
                    var precision = DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["precision"], Histogram.DefaultHistogramSettings.xAxisSettings.precision);
                    if (precision <= Histogram.MinPrecision) {
                        return Histogram.MinPrecision;
                    }
                    else if (precision >= Histogram.MaxPrecision) {
                        return Histogram.MaxPrecision;
                    }
                    return precision;
                };
                Histogram.getXAxisShow = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["show"], Histogram.DefaultHistogramSettings.xAxisSettings.show);
                };
                Histogram.getXAxisColor = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["axisColor"], {
                        solid: {
                            color: Histogram.DefaultHistogramSettings.xAxisSettings.axisColor
                        }
                    });
                };
                Histogram.getXTitle = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["xAxis"]["title"], Histogram.DefaultHistogramSettings.xAxisSettings.title);
                };
                Histogram.getYStyle = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["style"], Histogram.DefaultHistogramSettings.yAxisSettings.style);
                };
                Histogram.getYPosition = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["position"], Histogram.DefaultHistogramSettings.yAxisSettings.position);
                };
                Histogram.getYAxisShow = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["show"], Histogram.DefaultHistogramSettings.yAxisSettings.show);
                };
                Histogram.getYAxisColor = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["axisColor"], {
                        solid: {
                            color: Histogram.DefaultHistogramSettings.yAxisSettings.axisColor
                        }
                    });
                };
                Histogram.getYStart = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["start"], Histogram.DefaultHistogramSettings.yAxisSettings.start);
                };
                Histogram.getYEnd = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["end"], Histogram.DefaultHistogramSettings.yAxisSettings.end);
                };
                Histogram.getYDisplayUnit = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["displayUnits"], Histogram.DefaultHistogramSettings.yAxisSettings.displayUnits);
                };
                Histogram.getYPrecision = function (objects) {
                    var precision = DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["precision"], Histogram.DefaultHistogramSettings.yAxisSettings.precision);
                    if (precision <= Histogram.MinPrecision) {
                        return Histogram.MinPrecision;
                    }
                    else if (precision >= Histogram.MaxPrecision) {
                        return Histogram.MaxPrecision;
                    }
                    return precision;
                };
                Histogram.getYTitle = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["yAxis"]["title"], Histogram.DefaultHistogramSettings.yAxisSettings.title);
                };
                Histogram.getBins = function (objects) {
                    var binsNumber = Number(DataViewObjects.getValue(objects, Histogram.Properties["general"]["bins"], Histogram.DefaultHistogramSettings.bins));
                    if (!binsNumber || isNaN(binsNumber) || (binsNumber <= Histogram.MinNumberOfBins)) {
                        return Histogram.DefaultHistogramSettings.bins;
                    }
                    if (binsNumber > Histogram.MaxNumberOfBins) {
                        return Histogram.MaxNumberOfBins;
                    }
                    return binsNumber;
                };
                Histogram.getFrequency = function (objects) {
                    return DataViewObjects.getValue(objects, Histogram.Properties["general"]["frequency"], Histogram.DefaultHistogramSettings.frequency);
                };
                Histogram.getPrecision = function (objects) {
                    var precision = DataViewObjects.getValue(objects, Histogram.Properties["labels"]["precision"], Histogram.DefaultHistogramSettings.precision);
                    if (precision <= Histogram.MinPrecision) {
                        return Histogram.MinPrecision;
                    }
                    if (precision >= Histogram.MaxPrecision) {
                        return Histogram.MaxPrecision;
                    }
                    return precision;
                };
                Histogram.prototype.isDataValid = function (data) {
                    if (!data || !data.dataPoints || data.dataPoints.length === 0) {
                        return false;
                    }
                    if (data.dataPoints.some(function (x) { return x.range.some(function (x) { return isNaN(x) || x === Infinity || x === -Infinity; }); })) {
                        this.hostService.setWarnings([
                            new HistogramChartWarning(HistogramChartWarning.ErrorInvalidDataValues)
                        ]);
                        return false;
                    }
                    return true;
                };
                Histogram.prototype.update = function (visualUpdateOptions) {
                    if (!visualUpdateOptions ||
                        !visualUpdateOptions.dataViews ||
                        !visualUpdateOptions.dataViews[0]) {
                        return;
                    }
                    var dataView = visualUpdateOptions.dataViews[0], maxWidthOfVerticalAxisLabel;
                    this.setSize(visualUpdateOptions.viewport);
                    this.dataView = Histogram.converter(dataView, this.colors);
                    if (!this.isDataValid(this.dataView)) {
                        this.clear();
                        return;
                    }
                    this.updateViewportIn();
                    maxWidthOfVerticalAxisLabel = this.updateAxes(dataView);
                    this.columsAndAxesTransform(maxWidthOfVerticalAxisLabel);
                    this.updateWidthOfColumn();
                    this.createScales();
                    this.applySelectionStateToData();
                    this.render();
                };
                Histogram.prototype.updateAxes = function (dataView) {
                    var maxWidthOfVerticalAxisLabel, maxWidthOfHorizontalAxisLabel, maxHeightOfVerticalAxisLabel;
                    maxWidthOfVerticalAxisLabel = Histogram.getWidthOfLabel(this.dataView.borderValues.maxY, this.dataView.yLabelFormatter);
                    maxWidthOfHorizontalAxisLabel = Histogram.getWidthOfLabel(this.dataView.borderValues.maxX, this.dataView.xLabelFormatter);
                    maxHeightOfVerticalAxisLabel = Histogram.getHeightOfLabel(this.dataView.borderValues.maxX, this.dataView.xLabelFormatter);
                    var ySource = dataView.categorical.values &&
                        dataView.categorical.values[0] &&
                        dataView.categorical.values[0].values
                        ? dataView.categorical.values[0].source
                        : dataView.categorical.categories[0].source;
                    this.yAxisProperties = this.calculateYAxes(ySource, maxHeightOfVerticalAxisLabel);
                    this.renderYAxis();
                    this.yTitleMargin = this.shouldShowYOnRight()
                        ? this.viewport.width
                            - Histogram.YTitleMargin
                            + this.dataView.yLegendSize
                        : 0;
                    this.updateViewportIn(maxWidthOfVerticalAxisLabel);
                    this.xAxisProperties = this.calculateXAxes(dataView.categorical.categories[0].source, Histogram.DefaultTextProperties, maxWidthOfHorizontalAxisLabel, false);
                    this.renderXAxis();
                    return maxWidthOfVerticalAxisLabel;
                };
                Histogram.prototype.applySelectionStateToData = function () {
                    var _this = this;
                    if (this.interactivityService) {
                        this.dataView.dataPoints.forEach(function (dataPoint) {
                            _this.interactivityService.applySelectionStateToData(dataPoint.subDataPoints);
                        });
                    }
                };
                Histogram.prototype.createScales = function () {
                    var yAxisSettings = this.dataView.settings.yAxisSettings, borderValues = this.dataView.borderValues;
                    this.dataView.xScale = d3.scale.linear()
                        .domain([
                        borderValues.minX,
                        borderValues.maxX
                    ])
                        .range([
                        0,
                        this.viewportIn.width
                    ]);
                    this.dataView.yScale = d3.scale.linear()
                        .domain([
                        yAxisSettings.start,
                        yAxisSettings.end
                    ])
                        .range([
                        this.viewportIn.height,
                        this.outerPadding
                    ]);
                };
                Histogram.prototype.updateViewportIn = function (maxWidthOfVerticalAxisLabel) {
                    if (maxWidthOfVerticalAxisLabel === void 0) { maxWidthOfVerticalAxisLabel = 0; }
                    var width, height;
                    width = this.viewport.width
                        - this.dataView.yLegendSize
                        - maxWidthOfVerticalAxisLabel;
                    height = this.viewport.height - this.dataView.xLegendSize;
                    this.viewportIn = {
                        height: Math.max(height, Histogram.MinViewportInSize),
                        width: Math.max(width, Histogram.MinViewportInSize)
                    };
                };
                Histogram.prototype.updateWidthOfColumn = function () {
                    var countOfValues = this.dataView.dataPoints.length, widthOfColumn;
                    widthOfColumn = countOfValues
                        ? this.viewportIn.width / countOfValues - Histogram.ColumnPadding
                        : Histogram.MinViewportInSize;
                    this.widthOfColumn = Math.max(widthOfColumn, Histogram.MinViewportInSize);
                };
                Histogram.prototype.clear = function () {
                    var _this = this;
                    [
                        this.axisX,
                        this.axisY,
                        this.legend,
                        this.columns,
                        this.labelGraphicsContext
                    ].forEach(function (selection) {
                        _this.clearElement(selection);
                    });
                };
                Histogram.prototype.clearElement = function (selection) {
                    selection
                        .selectAll("*")
                        .remove();
                };
                Histogram.getLegendSize = function (axisSettings) {
                    return axisSettings.title
                        ? Histogram.LegendSizeWhenTitleIsActive
                        : Histogram.LegendSizeWhenTitleIsNotActive;
                };
                Histogram.getWidthOfLabel = function (labelValue, valueFormatter) {
                    var textProperties = Histogram.getTextPropertiesForMeasurement(labelValue, valueFormatter);
                    return TextMeasurementService.measureSvgTextWidth(textProperties) + Histogram.AdditionalWidthOfLabel;
                };
                Histogram.getHeightOfLabel = function (labelValue, valueFormatter) {
                    var textProperties = Histogram.getTextPropertiesForMeasurement(labelValue, valueFormatter);
                    return TextMeasurementService.measureSvgTextHeight(textProperties) + Histogram.AdditionalHeightOfLabel;
                };
                Histogram.getTextPropertiesForMeasurement = function (labelValue, valueFormatter) {
                    var labelText;
                    if (valueFormatter) {
                        labelText = valueFormatter.format(labelValue);
                    }
                    else {
                        labelText = labelValue;
                    }
                    return Histogram.getTextProperties(labelText);
                };
                Histogram.prototype.setSize = function (viewport) {
                    var height, width;
                    height = viewport.height -
                        this.margin.top -
                        this.margin.bottom;
                    width = viewport.width -
                        this.margin.left -
                        this.margin.right;
                    this.viewport = {
                        height: Math.max(height, Histogram.MinViewportSize),
                        width: Math.max(width, Histogram.MinViewportSize)
                    };
                    this.updateElements(Math.max(viewport.height, Histogram.MinViewportSize), Math.max(viewport.width, Histogram.MinViewportSize));
                };
                Histogram.prototype.updateElements = function (height, width) {
                    var transform = SVGUtil.translate(this.margin.left, this.margin.top);
                    this.root.attr({
                        height: height,
                        width: width
                    });
                    this.main.attr("transform", transform);
                    this.legend.attr("transform", transform);
                };
                Histogram.prototype.shouldShowYOnRight = function () {
                    return this.dataView.settings.yAxisSettings.position === yAxisPosition.right;
                };
                Histogram.prototype.columsAndAxesTransform = function (labelWidth) {
                    var offsetToRightStr, offsetToRight = this.shouldShowYOnRight()
                        ? this.margin.left
                        : this.dataView.settings.yAxisSettings.title
                            ? this.margin.left + labelWidth + Histogram.YAxisMargin
                            : this.margin.left + labelWidth;
                    offsetToRightStr = SVGUtil.translate(offsetToRight + Histogram.ColumnAndLabelOffset, 0);
                    this.columns.attr("transform", offsetToRightStr);
                    this.labelGraphicsContext.attr("transform", offsetToRightStr);
                    this.axes.attr("transform", SVGUtil.translate(offsetToRight, 0));
                    this.axisY.attr("transform", SVGUtil.translate(this.shouldShowYOnRight()
                        ? this.viewportIn.width : 0, 0));
                    this.axisX.attr("transform", SVGUtil.translate(0, this.viewportIn.height));
                };
                Histogram.prototype.render = function () {
                    var columnsSelection = this.renderColumns();
                    Histogram.bindTooltipsToSelection(columnsSelection);
                    this.bindSelectionHandler(columnsSelection);
                    this.renderLegend();
                    this.renderLabels();
                };
                Histogram.prototype.renderColumns = function () {
                    var _this = this;
                    var data = this.dataView.dataPoints, xScale = this.dataView.xScale, yScale = this.dataView.yScale, updateColumnsSelection;
                    updateColumnsSelection = this.columnsSelection.data(data);
                    updateColumnsSelection
                        .enter()
                        .append("svg:rect")
                        .classed(Histogram.Column.class, true);
                    updateColumnsSelection
                        .attr({
                        "x": function (dataPoint) {
                            return xScale(dataPoint.x);
                        },
                        "y": function (dataPoint) {
                            return yScale(dataPoint.y);
                        },
                        "width": this.widthOfColumn,
                        "height": function (item) {
                            return _this.getColumnHeight(item, yScale);
                        }
                    })
                        .style("fill", this.dataView.settings.fillColor);
                    histogramUtils.updateFillOpacity(updateColumnsSelection, this.interactivityService, false);
                    updateColumnsSelection
                        .exit()
                        .remove();
                    return updateColumnsSelection;
                };
                Histogram.bindTooltipsToSelection = function (selection) {
                    TooltipManager.addTooltip(selection, function (tooltipEvent) {
                        return tooltipEvent.data.tooltipInfo;
                    });
                };
                Histogram.prototype.getColumnHeight = function (column, y) {
                    var height = this.viewportIn.height - y(column.y);
                    return Math.max(height, Histogram.MinColumnHeight);
                };
                Histogram.prototype.renderXAxis = function () {
                    var _this = this;
                    var xAxis, xShow = this.dataView.settings.xAxisSettings.show, axisColor = this.dataView.settings.xAxisSettings.axisColor;
                    xAxis = this.xAxisProperties.axis
                        .tickFormat(function (value, index) {
                        var tickValues = _this.xAxisProperties.axis.tickValues(), amountOfLabels = (tickValues && tickValues.length) || 0;
                        return _this.formatLabelOfXAxis(value, index, amountOfLabels);
                    })
                        .orient("bottom");
                    if (xShow) {
                        this.axisX.call(xAxis);
                    }
                    else {
                        this.axisX
                            .selectAll("*")
                            .remove();
                    }
                    this.updateFillColorOfAxis(this.axisX, axisColor);
                };
                Histogram.prototype.formatLabelOfXAxis = function (labelValue, index, amountOfLabels) {
                    var maxWidthOfTheLatestLabel, formattedLabel = this.dataView.xLabelFormatter.format(labelValue);
                    if (index === 0 || index === amountOfLabels - 1) {
                        maxWidthOfTheLatestLabel = Math.min(this.viewportIn.width, Histogram.MaxWidthOfTheLatestLabel);
                        formattedLabel = Histogram.getTailoredTextOrDefault(formattedLabel, maxWidthOfTheLatestLabel);
                    }
                    return formattedLabel;
                };
                Histogram.getTailoredTextOrDefault = function (text, maxWidth) {
                    var textProperties = Histogram.getTextProperties(text);
                    return TextMeasurementService.getTailoredTextOrDefault(textProperties, maxWidth);
                };
                Histogram.getTextProperties = function (text) {
                    return {
                        text: text,
                        fontFamily: Histogram.DefaultTextProperties.fontFamily,
                        fontSize: Histogram.DefaultTextProperties.fontSize
                    };
                };
                Histogram.prototype.renderYAxis = function () {
                    var _this = this;
                    var yAxis, yShow = this.dataView.settings.yAxisSettings.show, axisColor = this.dataView.settings.yAxisSettings.axisColor;
                    yAxis = this.yAxisProperties.axis
                        .orient(this.dataView.settings.yAxisSettings.position.toLowerCase())
                        .tickFormat(function (item) {
                        return _this.dataView.yLabelFormatter.format(item);
                    });
                    if (yShow) {
                        this.axisY.call(yAxis);
                    }
                    else {
                        this.axisY
                            .selectAll("*")
                            .remove();
                    }
                    this.updateFillColorOfAxis(this.axisY, axisColor);
                };
                Histogram.prototype.updateFillColorOfAxis = function (axisSelection, fillColor) {
                    axisSelection
                        .selectAll("g.tick text")
                        .style({
                        "fill": fillColor
                    });
                };
                Histogram.prototype.getLabelLayout = function () {
                    var _this = this;
                    var labelSettings = this.dataView.settings.labelSettings, fontSizeInPx = PixelConverter.fromPoint(labelSettings.fontSize), fontFamily = dataLabelUtils.LabelTextProperties.fontFamily, xScale = this.dataView.xScale, yScale = this.dataView.yScale, dataLabelFormatter = ValueFormatter.create({
                        value: labelSettings.displayUnits,
                        precision: labelSettings.precision
                    });
                    return {
                        labelText: function (dataPoint) {
                            return dataLabelFormatter.format(dataPoint.y).toString();
                        },
                        labelLayout: {
                            x: function (dataPoint) {
                                var x, dx;
                                x = xScale(dataPoint.x);
                                dx = dataPoint.size.width / Histogram.DataLabelXOffset - _this.widthOfColumn / 2;
                                return x - dx;
                            },
                            y: function (dataPoint) {
                                var y, dy;
                                y = yScale(dataPoint.y);
                                dy = dataPoint.size.height;
                                return y - dy;
                            }
                        },
                        filter: function (dataPoint) {
                            return (dataPoint != null);
                        },
                        style: {
                            "fill": labelSettings.color,
                            "font-size": fontSizeInPx,
                            "font-family": fontFamily
                        }
                    };
                };
                Histogram.prototype.renderLabels = function () {
                    var labelSettings = this.dataView.settings.labelSettings, dataPointsArray = this.dataView.dataPoints, labels;
                    if (!labelSettings.show) {
                        dataLabelUtils.cleanDataLabels(this.labelGraphicsContext);
                        return;
                    }
                    labels = dataLabelUtils.drawDefaultLabelsForDataPointChart(dataPointsArray, this.labelGraphicsContext, this.getLabelLayout(), this.viewportIn);
                    if (labels) {
                        labels.attr("transform", function (dataPoint) {
                            var size = dataPoint.size, dx, dy;
                            dx = size.width / Histogram.DataLabelXOffset;
                            dy = size.height / Histogram.DataLabelYOffset;
                            return SVGUtil.translate(dx, dy);
                        });
                    }
                };
                Histogram.rangesToArray = function (data) {
                    return data.reduce(function (previousValue, currentValue, index) {
                        var range;
                        range = (index === 0)
                            ? currentValue.range
                            : currentValue.range.slice(1);
                        return previousValue.concat(range);
                    }, []);
                };
                Histogram.rangeToString = function (range, includeLeftBorder, valueFormatter) {
                    var leftBracket, rightBracket = Histogram.IncludeBrackets.right, leftBorder = valueFormatter.format(range[0]), rightBorder = valueFormatter.format(range[1]);
                    leftBracket = includeLeftBorder
                        ? Histogram.IncludeBrackets.left
                        : Histogram.ExcludeBrackets.left;
                    return "" + leftBracket + leftBorder + Histogram.SeparatorNumbers + rightBorder + rightBracket;
                };
                Histogram.prototype.renderLegend = function () {
                    var legendElements, legendSelection, datalegends = this.getDataLegends(this.dataView.settings);
                    legendElements = this.main
                        .select(Histogram.Legends.selector)
                        .selectAll(Histogram.Legend.selector);
                    legendSelection = legendElements.data(datalegends);
                    legendSelection
                        .enter()
                        .append("svg:text");
                    legendSelection
                        .attr({
                        "x": 0,
                        "y": 0,
                        "dx": function (item) { return item.dx; },
                        "dy": function (item) { return item.dy; },
                        "transform": function (item) { return item.transform; }
                    })
                        .text(function (item) { return item.text; })
                        .classed(Histogram.Legend.class, true);
                    legendSelection
                        .exit()
                        .remove();
                    this.legend
                        .select("text")
                        .style({
                        "display": Histogram.getDispayForAxisTitle(this.dataView.settings.xAxisSettings)
                    });
                    this.legend
                        .selectAll("text")
                        .filter(function (d, index) { return index === 1; })
                        .style({
                        "display": Histogram.getDispayForAxisTitle(this.dataView.settings.yAxisSettings)
                    });
                };
                Histogram.getDispayForAxisTitle = function (axisSettings) {
                    return axisSettings && axisSettings.title
                        ? null
                        : "none";
                };
                Histogram.prototype.getDataLegends = function (settings) {
                    var bottomLegendText = Histogram.getLegendText(settings);
                    bottomLegendText = Histogram.getLegend(bottomLegendText, settings.yAxisSettings.style, settings.yAxisSettings.displayUnits);
                    return [
                        {
                            transform: SVGUtil.translate(this.viewport.width / 2, this.viewport.height),
                            text: Histogram.getTailoredTextOrDefault(settings.displayName, this.viewportIn.width),
                            dx: "-0.5em",
                            dy: "-1em"
                        }, {
                            transform: SVGUtil.translateAndRotate(this.shouldShowYOnRight() ? this.yTitleMargin : 0, this.viewport.height / 2, 0, 0, 270),
                            text: Histogram.getTailoredTextOrDefault(bottomLegendText, this.viewportIn.height),
                            dx: "3em"
                        }
                    ];
                };
                Histogram.getLegendText = function (settings) {
                    return settings.frequency
                        ? Histogram.FrequencyText
                        : Histogram.DensityText;
                };
                Histogram.prototype.bindSelectionHandler = function (columnsSelection) {
                    if (!this.interactivityService
                        || !this.dataView
                        || !this.dataView.dataPoints) {
                        return;
                    }
                    var subDataPoints = [];
                    this.dataView.dataPoints.forEach(function (dataPoint) {
                        subDataPoints = subDataPoints.concat(dataPoint.subDataPoints);
                    });
                    var behaviorOptions = {
                        columns: columnsSelection,
                        clearCatcher: this.clearCatcher,
                        interactivityService: this.interactivityService,
                    };
                    this.interactivityService.bind(subDataPoints, this.behavior, behaviorOptions);
                };
                Histogram.prototype.enumerateObjectInstances = function (options) {
                    var enumeration = new ObjectEnumerationBuilder(), settings;
                    if (!this.dataView ||
                        !this.dataView.settings) {
                        return [];
                    }
                    settings = this.dataView.settings;
                    switch (options.objectName) {
                        case "general": {
                            this.enumerateGeneral(enumeration, settings);
                            break;
                        }
                        case "dataPoint": {
                            this.enumerateDataPoint(enumeration, settings);
                            break;
                        }
                        case "labels": {
                            this.enumerateLabels(enumeration, settings);
                            break;
                        }
                        case "xAxis": {
                            this.enumerateXAxis(enumeration, settings);
                            break;
                        }
                        case "yAxis": {
                            this.enumerateYAxis(enumeration, settings);
                            break;
                        }
                    }
                    return enumeration.complete() || [];
                };
                Histogram.prototype.enumerateGeneral = function (enumeration, settings) {
                    var general = {
                        objectName: "general",
                        displayName: "general",
                        selector: null,
                        properties: {
                            bins: settings.bins,
                            frequency: settings.frequency
                        }
                    };
                    enumeration.pushInstance(general);
                };
                Histogram.prototype.enumerateDataPoint = function (enumeration, settings) {
                    var dataPoint = {
                        objectName: "dataPoint",
                        displayName: "dataPoint",
                        selector: null,
                        properties: {
                            fill: settings.fillColor
                        }
                    };
                    enumeration.pushInstance(dataPoint);
                };
                Histogram.prototype.enumerateLabels = function (enumeration, settings) {
                    var labelsSettings = settings.labelSettings, labels = {
                        objectName: "labels",
                        displayName: "labels",
                        selector: null,
                        properties: {
                            show: labelsSettings.show,
                            color: labelsSettings.color,
                            displayUnits: labelsSettings.displayUnits,
                            precision: labelsSettings.precision,
                            fontSize: labelsSettings.fontSize
                        }
                    };
                    enumeration.pushInstance(labels);
                };
                Histogram.prototype.enumerateXAxis = function (enumeration, settings) {
                    var xAxisSettings = settings.xAxisSettings, xAxis = {
                        objectName: "xAxis",
                        displayName: "X-Axis",
                        selector: null,
                        properties: {
                            show: xAxisSettings.show,
                            title: xAxisSettings.title,
                            style: xAxisSettings.style,
                            axisColor: xAxisSettings.axisColor,
                            displayUnits: xAxisSettings.displayUnits,
                            precision: xAxisSettings.precision,
                        }
                    };
                    enumeration.pushInstance(xAxis);
                };
                Histogram.prototype.enumerateYAxis = function (enumeration, settings) {
                    var yAxisSettings = settings.yAxisSettings, yAxis = {
                        objectName: "yAxis",
                        displayName: "Y-Axis",
                        selector: null,
                        properties: {
                            show: yAxisSettings.show,
                            position: yAxisSettings.position,
                            start: yAxisSettings.start,
                            end: yAxisSettings.end,
                            title: yAxisSettings.title,
                            style: yAxisSettings.style,
                            axisColor: yAxisSettings.axisColor,
                            displayUnits: yAxisSettings.displayUnits,
                            precision: yAxisSettings.precision,
                        }
                    };
                    enumeration.pushInstance(yAxis);
                };
                Histogram.getObjectsFromDataView = function (dataView) {
                    if (!dataView ||
                        !dataView.metadata ||
                        !dataView.metadata.columns ||
                        !dataView.metadata.objects) {
                        return null;
                    }
                    return dataView.metadata.objects;
                };
                Histogram.prototype.destroy = function () {
                    this.root = null;
                };
                Histogram.prototype.calculateXAxes = function (source, textProperties, widthOfLabel, scrollbarVisible) {
                    var axes, width = this.viewportIn.width;
                    axes = this.calculateXAxesProperties(Histogram.rangesToArray(this.dataView.dataPoints), axisScale.linear, source, Histogram.InnerPaddingRatio, widthOfLabel);
                    axes.willLabelsFit = willLabelsFit(axes, width, TextMeasurementService.measureSvgTextWidth, textProperties);
                    // If labels do not fit and we are not scrolling, try word breaking
                    axes.willLabelsWordBreak = (!axes.willLabelsFit && !scrollbarVisible) && willLabelsWordBreak(axes, this.margin, width, TextMeasurementService.measureSvgTextWidth, TextMeasurementService.estimateSvgTextHeight, TextMeasurementService.getTailoredTextOrDefault, textProperties);
                    return axes;
                };
                Histogram.prototype.calculateXAxesProperties = function (forcedXDomain, categoryAxisScaleType, metaDataColumn, innerPaddingRatio, minOrdinalRectThickness) {
                    var xAxisProperties = HistogramAxisHelper.createAxis({
                        pixelSpan: this.viewportIn.width,
                        dataDomain: forcedXDomain,
                        metaDataColumn: metaDataColumn,
                        formatString: valueFormatter.getFormatString(metaDataColumn, Histogram.Properties["general"]["formatString"]),
                        outerPadding: 0,
                        isScalar: false,
                        isVertical: false,
                        useTickIntervalForDisplayUnits: true,
                        isCategoryAxis: true,
                        getValueFn: function (index, type) { return index; },
                        scaleType: categoryAxisScaleType,
                        innerPaddingRatio: innerPaddingRatio,
                        minOrdinalRectThickness: minOrdinalRectThickness,
                        tickLabelPadding: undefined
                    });
                    xAxisProperties.axisLabel = this.dataView.settings.displayName;
                    return xAxisProperties;
                };
                Histogram.prototype.calculateYAxes = function (source, heightOfLabel) {
                    var yAxisSettings = this.dataView.settings.yAxisSettings;
                    return this.calculateYAxesProperties([yAxisSettings.start, yAxisSettings.end], axisScale.linear, source, Histogram.InnerPaddingRatio, heightOfLabel);
                };
                Histogram.prototype.calculateYAxesProperties = function (forcedYDomain, categoryAxisScaleType, metaDataColumn, innerPaddingRatio, minOrdinalRectThickness) {
                    var formatString = undefined;
                    if (this.dataView.settings.frequency) {
                        formatString = valueFormatter.getFormatString(metaDataColumn, Histogram.Properties["general"]["formatString"]);
                    }
                    return HistogramAxisHelper.createAxis({
                        pixelSpan: this.viewportIn.height,
                        dataDomain: forcedYDomain,
                        metaDataColumn: metaDataColumn,
                        formatString: formatString,
                        outerPadding: this.outerPadding,
                        isScalar: true,
                        isVertical: true,
                        useTickIntervalForDisplayUnits: true,
                        isCategoryAxis: false,
                        getValueFn: function (index, type) { return index; },
                        scaleType: categoryAxisScaleType,
                        innerPaddingRatio: innerPaddingRatio,
                        minOrdinalRectThickness: minOrdinalRectThickness,
                        tickLabelPadding: undefined,
                        is100Pct: true
                    });
                };
                Histogram.ClassName = "histogram";
                Histogram.FrequencyText = "Frequency";
                Histogram.DensityText = "Density";
                Histogram.Properties = {
                    general: {
                        bins: {
                            objectName: "general",
                            propertyName: "bins"
                        },
                        frequency: {
                            objectName: "general",
                            propertyName: "frequency"
                        },
                        formatString: {
                            objectName: "general",
                            propertyName: "formatString"
                        }
                    },
                    dataPoint: {
                        fill: {
                            objectName: "dataPoint",
                            propertyName: "fill"
                        }
                    },
                    labels: {
                        show: {
                            objectName: "labels",
                            propertyName: "show"
                        },
                        color: {
                            objectName: "labels",
                            propertyName: "color"
                        },
                        displayUnits: {
                            objectName: "labels",
                            propertyName: "displayUnits"
                        },
                        precision: {
                            objectName: "labels",
                            propertyName: "precision"
                        },
                        fontSize: {
                            objectName: "labels",
                            propertyName: "fontSize"
                        }
                    },
                    xAxis: {
                        show: {
                            objectName: "xAxis",
                            propertyName: "show"
                        },
                        axisColor: {
                            objectName: "xAxis",
                            propertyName: "axisColor"
                        },
                        title: {
                            objectName: "xAxis",
                            propertyName: "title"
                        },
                        displayUnits: {
                            objectName: "xAxis",
                            propertyName: "displayUnits"
                        },
                        precision: {
                            objectName: "xAxis",
                            propertyName: "precision"
                        },
                        style: {
                            objectName: "xAxis",
                            propertyName: "style"
                        }
                    },
                    yAxis: {
                        show: {
                            objectName: "yAxis",
                            propertyName: "show"
                        },
                        axisColor: {
                            objectName: "yAxis",
                            propertyName: "axisColor"
                        },
                        title: {
                            objectName: "yAxis",
                            propertyName: "title"
                        },
                        displayUnits: {
                            objectName: "yAxis",
                            propertyName: "displayUnits"
                        },
                        precision: {
                            objectName: "yAxis",
                            propertyName: "precision"
                        },
                        style: {
                            objectName: "yAxis",
                            propertyName: "style"
                        },
                        start: {
                            objectName: "yAxis",
                            propertyName: "start"
                        },
                        end: {
                            objectName: "yAxis",
                            propertyName: "end"
                        },
                        position: {
                            objectName: "yAxis",
                            propertyName: "position"
                        }
                    }
                };
                Histogram.DefaultHistogramSettings = {
                    frequency: true,
                    displayName: "Histogram",
                    bins: null,
                    fillColor: "#01b8aa",
                    precision: 2,
                    xAxisSettings: {
                        show: true,
                        axisColor: "#777",
                        title: true,
                        displayUnits: 0,
                        precision: 2,
                        style: axisStyle.showTitleOnly,
                    },
                    yAxisSettings: {
                        show: true,
                        axisColor: "#777",
                        title: true,
                        displayUnits: 0,
                        precision: 2,
                        style: axisStyle.showTitleOnly,
                        start: 0,
                        position: yAxisPosition.left,
                    },
                    labelSettings: {
                        show: false,
                        color: "#777",
                        displayUnits: 0,
                        precision: 2,
                        fontSize: 9
                    },
                };
                Histogram.Axes = createClassAndSelector("axes");
                Histogram.Axis = createClassAndSelector("axis");
                Histogram.XAxis = createClassAndSelector("xAxis");
                Histogram.YAxis = createClassAndSelector("yAxis");
                Histogram.Columns = createClassAndSelector("columns");
                Histogram.Column = createClassAndSelector("column");
                Histogram.Legends = createClassAndSelector("legends");
                Histogram.Legend = createClassAndSelector("legend");
                Histogram.LabelGraphicsContext = createClassAndSelector("labelGraphicsContext");
                Histogram.MinNumberOfBins = 0;
                Histogram.MaxNumberOfBins = 100;
                Histogram.MinPrecision = 0;
                Histogram.MaxPrecision = 17; // max number of decimals in float
                Histogram.MinXAxisStartValue = 0;
                Histogram.MaxXAxisEndValue = 1e+25;
                Histogram.YTitleMargin = 70;
                Histogram.YAxisMargin = 20;
                Histogram.MinViewportSize = 100;
                Histogram.MinViewportInSize = 0;
                Histogram.MinAmountOfValues = 2;
                Histogram.AdditionalWidthOfLabel = 3;
                Histogram.AdditionalHeightOfLabel = 3;
                Histogram.LegendSizeWhenTitleIsActive = 50;
                Histogram.LegendSizeWhenTitleIsNotActive = 25;
                Histogram.InnerPaddingRatio = 1;
                Histogram.DataLabelXOffset = 2;
                Histogram.DataLabelYOffset = 1.8;
                Histogram.ColumnPadding = 2.5;
                Histogram.ColumnAndLabelOffset = 1.5;
                Histogram.MinColumnHeight = 1;
                Histogram.TooltipDisplayName = "Range";
                Histogram.SeparatorNumbers = ", ";
                Histogram.MaxWidthOfTheLatestLabel = 40;
                Histogram.ExcludeBrackets = {
                    left: "(",
                    right: ")"
                };
                Histogram.IncludeBrackets = {
                    left: "[",
                    right: "]"
                };
                Histogram.DefaultMargin = {
                    top: 10,
                    right: 10,
                    bottom: 10,
                    left: 10
                };
                Histogram.DefaultTextProperties = {
                    fontFamily: "'Segoe UI', 'wf_segoe-ui_normal', helvetica, arial, sans-serif",
                    fontSize: PixelConverter.toString(11) // Note: This value and font-size in histogram.less should be the same.
                };
                Histogram.capabilities = {
                    dataRoles: [
                        {
                            name: "Values",
                            kind: VisualDataRoleKind.Grouping,
                            displayName: "Values"
                        }, {
                            name: "Frequency",
                            kind: VisualDataRoleKind.Measure,
                            displayName: "Frequency"
                        }
                    ],
                    dataViewMappings: [{
                            conditions: [{ "Values": { min: 1, max: 1 }, "Frequency": { min: 0, max: 1 } }],
                            categorical: {
                                categories: {
                                    bind: { to: "Values" },
                                    dataReductionAlgorithm: { top: {} }
                                },
                                values: { for: { in: "Frequency" } }
                            }
                        }],
                    sorting: {
                        implicit: {
                            clauses: [{ role: "Values", direction: 1 /*SortDirection.Ascending*/ }] //Constant SortDirection.Ascending currently is not supported on the msit
                        }
                    },
                    objects: {
                        general: {
                            displayName: "General",
                            properties: {
                                formatString: { type: { formatting: { formatString: true } } },
                                bins: {
                                    displayName: "Bins",
                                    type: { numeric: true }
                                },
                                frequency: {
                                    displayName: "Frequency",
                                    type: { bool: true }
                                }
                            },
                        },
                        dataPoint: {
                            displayName: "Data colors",
                            properties: {
                                fill: {
                                    displayName: "Fill",
                                    type: { fill: { solid: { color: true } } }
                                }
                            }
                        },
                        xAxis: {
                            displayName: "X-Axis",
                            properties: {
                                show: {
                                    displayName: "Show",
                                    type: { bool: true },
                                },
                                axis: {
                                    displayName: "Axis",
                                    type: { bool: true }
                                },
                                axisColor: {
                                    displayName: "Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                title: {
                                    displayName: "Title",
                                    type: { bool: true }
                                },
                                displayUnits: {
                                    displayName: "Display Units",
                                    type: { formatting: { labelDisplayUnits: true } }
                                },
                                precision: {
                                    displayName: "Decimal Places",
                                    type: { numeric: true },
                                },
                                style: {
                                    displayName: "Style",
                                    type: { enumeration: axisStyle.type }
                                },
                            }
                        },
                        yAxis: {
                            displayName: "Y-Axis",
                            properties: {
                                show: {
                                    displayName: "Show",
                                    type: { bool: true },
                                },
                                axis: {
                                    displayName: "yAxis",
                                    type: { bool: true }
                                },
                                axisColor: {
                                    displayName: "Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                title: {
                                    displayName: "Title",
                                    type: { bool: true }
                                },
                                displayUnits: {
                                    displayName: "Display Units",
                                    type: { formatting: { labelDisplayUnits: true } }
                                },
                                precision: {
                                    displayName: "Decimal Places",
                                    type: { numeric: true },
                                },
                                style: {
                                    displayName: "Style",
                                    type: { enumeration: axisStyle.type }
                                },
                                start: {
                                    displayName: "Start",
                                    type: { numeric: true },
                                    placeHolderText: "Start",
                                    suppressFormatPainterCopy: true,
                                },
                                end: {
                                    displayName: "End",
                                    type: { numeric: true },
                                    placeHolderText: "End",
                                    suppressFormatPainterCopy: true,
                                },
                                position: {
                                    displayName: "Position",
                                    type: { enumeration: yAxisPosition.type },
                                },
                            }
                        },
                        labels: {
                            displayName: "Data Labels",
                            properties: {
                                show: {
                                    displayName: "Show",
                                    type: { bool: true }
                                },
                                color: {
                                    displayName: "Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                displayUnits: {
                                    displayName: "Display Units",
                                    type: { formatting: { labelDisplayUnits: true } },
                                    suppressFormatPainterCopy: true
                                },
                                precision: {
                                    displayName: "Decimal Places",
                                    type: { numeric: true },
                                    suppressFormatPainterCopy: true
                                },
                                fontSize: {
                                    displayName: "Text Size",
                                    type: { formatting: { fontSize: true } }
                                },
                            },
                        },
                    }
                };
                return Histogram;
            }());
            Histogram1445664487616.Histogram = Histogram;
            /**
             * HistogramAxisHelper based on AxisHelper (Visuals/common/axisHelper.ts).
             */
            var HistogramAxisHelper;
            (function (HistogramAxisHelper) {
                var NumberFormat = powerbi.NumberFormat;
                /**
                 * Default ranges are for when we have a field chosen for the axis,
                 * but no values are returned by the query.
                 */
                HistogramAxisHelper.emptyDomain = [0, 0];
                var InnerPaddingRatio = 0.2;
                var TickLabelPadding = 2; // between text labels, used by AxisHelper
                var MinOrdinalRectThickness = 20;
                var ScalarTickLabelPadding = 3;
                var MinTickCount = 2;
                var DefaultBestTickCount = 3;
                /**
                 * Create a D3 axis including scale. Can be vertical or horizontal, and either datetime, numeric, or text.
                 * @param options The properties used to create the axis.
                 */
                function createAxis(options) {
                    var pixelSpan = options.pixelSpan, dataDomain = options.dataDomain, metaDataColumn = options.metaDataColumn, formatString = options.formatString, outerPadding = options.outerPadding || 0, isCategoryAxis = !!options.isCategoryAxis, isScalar = !!options.isScalar, isVertical = !!options.isVertical, useTickIntervalForDisplayUnits = !!options.useTickIntervalForDisplayUnits, // DEPRECATE: same meaning as isScalar?
                    getValueFn = options.getValueFn, categoryThickness = options.categoryThickness, axisDisplayUnits = options.axisDisplayUnits, axisPrecision = options.axisPrecision, is100Pct = !!options.is100Pct, tickLabelPadding = options.tickLabelPadding || TickLabelPadding;
                    var dataType = getCategoryValueType(metaDataColumn, isScalar);
                    // Create the Scale
                    var scaleResult = createScale(options);
                    var scale = scaleResult.scale;
                    var bestTickCount = scaleResult.bestTickCount;
                    var scaleDomain = scale.domain();
                    var isLogScaleAllowed = isLogScalePossible(dataDomain, dataType);
                    // fix categoryThickness if scalar and the domain was adjusted when making the scale "nice"
                    if (categoryThickness && isScalar && dataDomain && dataDomain.length === 2) {
                        var oldSpan = dataDomain[1] - dataDomain[0];
                        var newSpan = scaleDomain[1] - scaleDomain[0];
                        if (oldSpan > 0 && newSpan > 0) {
                            categoryThickness = categoryThickness * oldSpan / newSpan;
                        }
                    }
                    // Prepare Tick Values for formatting
                    var tickValues;
                    if (isScalar && bestTickCount === 1) {
                        tickValues = [dataDomain[0]];
                    }
                    else {
                        var minTickInterval = isScalar ? getMinTickValueInterval(formatString, dataType, is100Pct) : undefined;
                        tickValues = getRecommendedTickValues(bestTickCount, scale, dataType, isScalar, minTickInterval);
                    }
                    if (options.scaleType && options.scaleType === axisScale.log && isLogScaleAllowed) {
                        tickValues = tickValues.filter(function (d) { return powerOfTen(d); });
                    }
                    var formatter = createFormatter(scaleDomain, dataDomain, dataType, isScalar, formatString, bestTickCount, tickValues, getValueFn, useTickIntervalForDisplayUnits, axisDisplayUnits, axisPrecision);
                    // sets default orientation only, cartesianChart will fix y2 for comboChart
                    // tickSize(pixelSpan) is used to create gridLines
                    var axis = d3.svg.axis()
                        .scale(scale)
                        .tickSize(6, 0)
                        .orient(isVertical ? "left" : "bottom")
                        .ticks(bestTickCount)
                        .tickValues(tickValues);
                    var formattedTickValues = [];
                    if (metaDataColumn)
                        formattedTickValues = formatAxisTickValues(axis, tickValues, formatter, dataType, getValueFn);
                    var xLabelMaxWidth;
                    // Use category layout of labels if specified, otherwise use scalar layout of labels
                    if (!isScalar && categoryThickness) {
                        xLabelMaxWidth = Math.max(1, categoryThickness - tickLabelPadding * 2);
                    }
                    else {
                        // When there are 0 or 1 ticks, then xLabelMaxWidth = pixelSpan
                        xLabelMaxWidth = tickValues.length > 1 ? getScalarLabelMaxWidth(scale, tickValues) : pixelSpan;
                        xLabelMaxWidth = xLabelMaxWidth - ScalarTickLabelPadding * 2;
                    }
                    return {
                        scale: scale,
                        axis: axis,
                        formatter: formatter,
                        values: formattedTickValues,
                        axisType: dataType,
                        axisLabel: null,
                        isCategoryAxis: isCategoryAxis,
                        xLabelMaxWidth: xLabelMaxWidth,
                        categoryThickness: categoryThickness,
                        outerPadding: outerPadding,
                        usingDefaultDomain: scaleResult.usingDefaultDomain,
                        isLogScaleAllowed: isLogScaleAllowed,
                        dataDomain: dataDomain,
                    };
                }
                HistogramAxisHelper.createAxis = createAxis;
                /**
                 * Indicates whether the number is power of 10.
                 */
                function powerOfTen(d) {
                    var value = Math.abs(d);
                    // formula log2(Y)/log2(10) = log10(Y)
                    // because double issues this won"t return exact value
                    // we need to ceil it to nearest number.
                    var log10 = Math.log(value) / Math.LN10;
                    log10 = Math.ceil(log10 - 1e-12);
                    return value / Math.pow(10, log10) === 1;
                }
                HistogramAxisHelper.powerOfTen = powerOfTen;
                function getScalarLabelMaxWidth(scale, tickValues) {
                    debug.assertValue(scale, "scale");
                    debug.assertNonEmpty(tickValues, "tickValues");
                    // find the distance between two ticks. scalar ticks can be anywhere, such as:
                    // |---50----------100--------|
                    if (scale && !_.isEmpty(tickValues)) {
                        return Math.abs(scale(tickValues[1]) - scale(tickValues[0]));
                    }
                    return 1;
                }
                function createFormatter(scaleDomain, dataDomain, dataType, isScalar, formatString, bestTickCount, tickValues, getValueFn, useTickIntervalForDisplayUnits, axisDisplayUnits, axisPrecision) {
                    if (useTickIntervalForDisplayUnits === void 0) { useTickIntervalForDisplayUnits = false; }
                    var formatter;
                    if (dataType.dateTime) {
                        if (isScalar) {
                            var value = new Date(scaleDomain[0]);
                            var value2 = new Date(scaleDomain[1]);
                            // datetime with only one value needs to pass the same value
                            // (from the original dataDomain value, not the adjusted scaleDomain)
                            // so formatting works correctly.
                            if (bestTickCount === 1)
                                value = value2 = new Date(dataDomain[0]);
                            // this will ignore the formatString and create one based on the smallest non-zero portion of the values supplied.
                            formatter = valueFormatter.create({
                                format: formatString,
                                value: value,
                                value2: value2,
                                tickCount: bestTickCount,
                            });
                        }
                        else {
                            // Use the model formatString for ordinal datetime
                            formatter = valueFormatter.createDefaultFormatter(formatString, true);
                        }
                    }
                    else {
                        if (getValueFn == null && !isScalar) {
                            debug.assertFail("getValueFn must be supplied for ordinal tickValues");
                        }
                        if (useTickIntervalForDisplayUnits && isScalar && tickValues.length > 1) {
                            var value1 = axisDisplayUnits ? axisDisplayUnits : tickValues[1] - tickValues[0];
                            var options = {
                                format: formatString,
                                value: value1,
                                value2: 0,
                                allowFormatBeautification: true,
                            };
                            if (axisPrecision)
                                options.precision = axisPrecision;
                            else
                                options.detectAxisPrecision = true;
                            formatter = valueFormatter.create(options);
                        }
                        else {
                            // do not use display units, just the basic value formatter
                            // datetime is handled above, so we are ordinal and either boolean, numeric, or text.
                            formatter = valueFormatter.createDefaultFormatter(formatString, true);
                        }
                    }
                    return formatter;
                }
                HistogramAxisHelper.createFormatter = createFormatter;
                function getMinTickValueInterval(formatString, columnType, is100Pct) {
                    var isCustomFormat = formatString && !NumberFormat.isStandardFormat(formatString);
                    if (isCustomFormat) {
                        var precision = NumberFormat.getCustomFormatMetadata(formatString, true /*calculatePrecision*/).precision;
                        if (formatString.indexOf("%") > -1)
                            precision += 2; //percent values are multiplied by 100 during formatting
                        return Math.pow(10, -precision);
                    }
                    else if (is100Pct)
                        return 0.01;
                    else if (columnType.integer)
                        return 1;
                    return 0;
                }
                HistogramAxisHelper.getMinTickValueInterval = getMinTickValueInterval;
                /**
                 * Format the linear tick labels or the category labels.
                 */
                function formatAxisTickValues(axis, tickValues, formatter, dataType, getValueFn) {
                    var formattedTickValues = [];
                    if (!getValueFn)
                        getValueFn = function (data) { return data; };
                    if (formatter) {
                        axis.tickFormat(function (d) { return formatter.format(getValueFn(d, dataType)); });
                        formattedTickValues = tickValues.map(function (d) { return formatter.format(getValueFn(d, dataType)); });
                    }
                    else {
                        formattedTickValues = tickValues.map(function (d) { return getValueFn(d, dataType); });
                    }
                    return formattedTickValues;
                }
                function isLogScalePossible(domain, axisType) {
                    if (domain == null)
                        return false;
                    if (isDateTime(axisType))
                        return false;
                    return (domain[0] > 0 && domain[1] > 0) || (domain[0] < 0 && domain[1] < 0); //doman must exclude 0
                }
                HistogramAxisHelper.isLogScalePossible = isLogScalePossible;
                function isDateTime(type) {
                    return !!(type && type.dateTime);
                }
                HistogramAxisHelper.isDateTime = isDateTime;
                function getRecommendedTickValues(maxTicks, scale, axisType, isScalar, minTickInterval) {
                    if (!isScalar || isOrdinalScale(scale)) {
                        return getRecommendedTickValuesForAnOrdinalRange(maxTicks, scale.domain());
                    }
                    else if (isDateTime(axisType)) {
                        return getRecommendedTickValuesForADateTimeRange(maxTicks, scale.domain());
                    }
                    return getRecommendedTickValuesForAQuantitativeRange(maxTicks, scale, minTickInterval);
                }
                HistogramAxisHelper.getRecommendedTickValues = getRecommendedTickValues;
                function getRecommendedTickValuesForAnOrdinalRange(maxTicks, labels) {
                    var tickLabels = [];
                    // return no ticks in this case
                    if (maxTicks <= 0)
                        return tickLabels;
                    var len = labels.length;
                    if (maxTicks > len)
                        return labels;
                    for (var i = 0, step = Math.ceil(len / maxTicks); i < len; i += step) {
                        tickLabels.push(labels[i]);
                    }
                    return tickLabels;
                }
                HistogramAxisHelper.getRecommendedTickValuesForAnOrdinalRange = getRecommendedTickValuesForAnOrdinalRange;
                function getRecommendedTickValuesForAQuantitativeRange(maxTicks, scale, minInterval) {
                    var tickLabels = [];
                    //if maxticks is zero return none
                    if (maxTicks === 0)
                        return tickLabels;
                    var quantitiveScale = scale;
                    if (quantitiveScale.ticks) {
                        tickLabels = quantitiveScale.ticks(maxTicks);
                        if (tickLabels.length > maxTicks && maxTicks > 1)
                            tickLabels = quantitiveScale.ticks(maxTicks - 1);
                        if (tickLabels.length < MinTickCount) {
                            tickLabels = quantitiveScale.ticks(maxTicks + 1);
                        }
                        tickLabels = createTrueZeroTickLabel(tickLabels);
                        if (minInterval && tickLabels.length > 1) {
                            var tickInterval = tickLabels[1] - tickLabels[0];
                            while (tickInterval > 0 && tickInterval < minInterval) {
                                for (var i = 1; i < tickLabels.length; i++) {
                                    tickLabels.splice(i, 1);
                                }
                                tickInterval = tickInterval * 2;
                            }
                            // keep at least two labels - the loop above may trim all but one if we have odd # of tick labels and dynamic range < minInterval
                            if (tickLabels.length === 1) {
                                tickLabels.push(tickLabels[0] + minInterval);
                            }
                        }
                        return tickLabels;
                    }
                    debug.assertFail("must pass a quantitative scale to this method");
                    return tickLabels;
                }
                HistogramAxisHelper.getRecommendedTickValuesForAQuantitativeRange = getRecommendedTickValuesForAQuantitativeRange;
                function getRecommendedTickValuesForADateTimeRange(maxTicks, dataDomain) {
                    var tickLabels = [];
                    if (dataDomain[0] === 0 && dataDomain[1] === 0)
                        return [];
                    var dateTimeTickLabels = DateTimeSequence.calculate(new Date(dataDomain[0]), new Date(dataDomain[1]), maxTicks).sequence;
                    tickLabels = dateTimeTickLabels.map(function (d) { return d.getTime(); });
                    tickLabels = ensureValuesInRange(tickLabels, dataDomain[0], dataDomain[1]);
                    return tickLabels;
                }
                function isOrdinalScale(scale) {
                    return typeof scale.invert === "undefined";
                }
                HistogramAxisHelper.isOrdinalScale = isOrdinalScale;
                /**
                 * Gets the ValueType of a category column, defaults to Text if the type is not present.
                 */
                function getCategoryValueType(metadataColumn, isScalar) {
                    if (metadataColumn && columnDataTypeHasValue(metadataColumn.type))
                        return metadataColumn.type;
                    if (isScalar) {
                        return ValueType.fromDescriptor({ numeric: true });
                    }
                    return ValueType.fromDescriptor({ text: true });
                }
                HistogramAxisHelper.getCategoryValueType = getCategoryValueType;
                function columnDataTypeHasValue(dataType) {
                    return dataType && (dataType.bool || dataType.numeric || dataType.text || dataType.dateTime);
                }
                HistogramAxisHelper.columnDataTypeHasValue = columnDataTypeHasValue;
                function createScale(options) {
                    var pixelSpan = options.pixelSpan, dataDomain = options.dataDomain, metaDataColumn = options.metaDataColumn, outerPadding = options.outerPadding || 0, isScalar = !!options.isScalar, isVertical = !!options.isVertical, forcedTickCount = options.forcedTickCount, categoryThickness = options.categoryThickness, shouldClamp = !!options.shouldClamp, maxTickCount = options.maxTickCount, innerPaddingRatio = options.innerPaddingRatio || InnerPaddingRatio, minOrdinalRectThickness = options.minOrdinalRectThickness || MinOrdinalRectThickness;
                    var dataType = getCategoryValueType(metaDataColumn, isScalar);
                    var maxTicks = isVertical
                        ? getRecommendedNumberOfTicksForYAxis(pixelSpan)
                        : getRecommendedNumberOfTicksForXAxis(pixelSpan);
                    if (maxTickCount &&
                        maxTicks > maxTickCount)
                        maxTicks = maxTickCount;
                    var scalarDomain = dataDomain ? dataDomain.slice() : null;
                    var bestTickCount = maxTicks;
                    var scale;
                    var usingDefaultDomain = false;
                    if (dataDomain == null || (dataDomain.length === 2 && dataDomain[0] == null && dataDomain[1] == null) || (dataDomain.length !== 2 && isScalar)) {
                        usingDefaultDomain = true;
                        if (dataType.dateTime || !isOrdinal(dataType))
                            dataDomain = HistogramAxisHelper.emptyDomain;
                        else
                            dataDomain = [];
                        if (isOrdinal(dataType)) {
                            scale = createOrdinalScale(pixelSpan, dataDomain, innerPaddingRatio, categoryThickness ? outerPadding / categoryThickness : 0);
                        }
                        else {
                            scale = createNumericalScale(options.scaleType, pixelSpan, dataDomain, dataType, outerPadding, bestTickCount);
                        }
                    }
                    else {
                        if (isScalar && dataDomain.length > 0) {
                            bestTickCount = forcedTickCount !== undefined
                                ? (maxTicks !== 0 ? forcedTickCount : 0)
                                : getBestNumberOfTicks(dataDomain[0], dataDomain[dataDomain.length - 1], [metaDataColumn], maxTicks, dataType.dateTime);
                            var normalizedRange = normalizeLinearDomain({ min: dataDomain[0], max: dataDomain[dataDomain.length - 1] });
                            scalarDomain = [normalizedRange.min, normalizedRange.max];
                        }
                        if (isScalar && dataType.numeric && !dataType.dateTime) {
                            //Note: Don't pass bestTickCount to createNumericalScale, because it overrides boundaries of the domain.
                            scale = createNumericalScale(options.scaleType, pixelSpan, scalarDomain, dataType, outerPadding, /*bestTickCount*/ null, shouldClamp);
                            bestTickCount = maxTicks === 0
                                ? 0
                                : Math.floor((pixelSpan - outerPadding * 2) / minOrdinalRectThickness);
                        }
                        else if (isScalar && dataType.dateTime) {
                            // Use of a linear scale, instead of a D3.time.scale, is intentional since we want
                            // to control the formatting of the time values, since d3"s implementation isn"t
                            // in accordance to our design.
                            //     scalarDomain: should already be in long-int time (via category.values[0].getTime())
                            scale = createLinearScale(pixelSpan, scalarDomain, outerPadding, null, shouldClamp); // DO NOT PASS TICKCOUNT
                        }
                        else if (dataType.text || dataType.dateTime || dataType.numeric || dataType.bool) {
                            scale = createOrdinalScale(pixelSpan, scalarDomain, innerPaddingRatio, categoryThickness ? outerPadding / categoryThickness : 0);
                            bestTickCount = maxTicks === 0 ? 0
                                : Math.min(scalarDomain.length, (pixelSpan - outerPadding * 2) / minOrdinalRectThickness);
                        }
                        else {
                            debug.assertFail("unsupported dataType, something other than text or numeric");
                        }
                    }
                    // vertical ordinal axis (e.g. categorical bar chart) does not need to reverse
                    if (isVertical && isScalar) {
                        scale.range(scale.range().reverse());
                    }
                    normalizeInfinityInScale(scale);
                    return {
                        scale: scale,
                        bestTickCount: bestTickCount,
                        usingDefaultDomain: usingDefaultDomain,
                    };
                }
                HistogramAxisHelper.createScale = createScale;
                function normalizeInfinityInScale(scale) {
                    // When large values (eg Number.MAX_VALUE) are involved, a call to scale.nice occasionally
                    // results in infinite values being included in the domain. To correct for that, we need to
                    // re-normalize the domain now to not include infinities.
                    var scaledDomain = scale.domain();
                    for (var i = 0, len = scaledDomain.length; i < len; ++i) {
                        if (scaledDomain[i] === Number.POSITIVE_INFINITY)
                            scaledDomain[i] = Number.MAX_VALUE;
                        else if (scaledDomain[i] === Number.NEGATIVE_INFINITY)
                            scaledDomain[i] = -Number.MAX_VALUE;
                    }
                    scale.domain(scaledDomain);
                }
                HistogramAxisHelper.normalizeInfinityInScale = normalizeInfinityInScale;
                function createOrdinalScale(pixelSpan, dataDomain, innerPaddingRatio, outerPaddingRatio) {
                    debug.assert(outerPaddingRatio >= 0 && outerPaddingRatio < 4, "outerPaddingRatio should be a value between zero and four");
                    var scale = d3.scale.ordinal()
                        .rangeBands([0, pixelSpan], innerPaddingRatio, outerPaddingRatio)
                        .domain(dataDomain);
                    return scale;
                }
                HistogramAxisHelper.createOrdinalScale = createOrdinalScale;
                function normalizeLinearDomain(domain) {
                    if (isNaN(domain.min) || isNaN(domain.max)) {
                        domain.min = HistogramAxisHelper.emptyDomain[0];
                        domain.max = HistogramAxisHelper.emptyDomain[1];
                    }
                    else if (domain.min === domain.max) {
                        // d3 linear scale will give zero tickValues if max === min, so extend a little
                        domain.min = domain.min < 0 ? domain.min * 1.2 : domain.min * 0.8;
                        domain.max = domain.max < 0 ? domain.max * 0.8 : domain.max * 1.2;
                    }
                    else {
                        // Check that min is very small and is a negligable portion of the whole domain.
                        // (fix floating pt precision bugs)
                        // sometimes highlight value math causes small negative numbers which makes the axis add
                        // a large tick interval instead of just rendering at zero.
                        if (Math.abs(domain.min) < 0.0001 && domain.min / (domain.max - domain.min) < 0.0001) {
                            domain.min = 0;
                        }
                    }
                    return domain;
                }
                //this function can return different scales e.g. log, linear
                // NOTE: export only for testing, do not access directly
                function createNumericalScale(axisScaleType, pixelSpan, dataDomain, dataType, outerPadding, niceCount, shouldClamp) {
                    if (outerPadding === void 0) { outerPadding = 0; }
                    if (axisScaleType === axisScale.log && isLogScalePossible(dataDomain, dataType)) {
                        return createLogScale(pixelSpan, dataDomain, outerPadding, niceCount);
                    }
                    else {
                        return createLinearScale(pixelSpan, dataDomain, outerPadding, niceCount, shouldClamp);
                    }
                }
                HistogramAxisHelper.createNumericalScale = createNumericalScale;
                function createLogScale(pixelSpan, dataDomain, outerPadding, niceCount) {
                    if (outerPadding === void 0) { outerPadding = 0; }
                    debug.assert(isLogScalePossible(dataDomain), "dataDomain cannot include 0");
                    var scale = d3.scale.log()
                        .range([outerPadding, pixelSpan - outerPadding])
                        .domain([dataDomain[0], dataDomain[1]])
                        .clamp(true);
                    if (niceCount) {
                        scale.nice(niceCount);
                    }
                    return scale;
                }
                // NOTE: export only for testing, do not access directly
                function createLinearScale(pixelSpan, dataDomain, outerPadding, niceCount, shouldClamp) {
                    if (outerPadding === void 0) { outerPadding = 0; }
                    var scale = d3.scale.linear()
                        .range([outerPadding, pixelSpan - outerPadding])
                        .domain([dataDomain[0], dataDomain[1]])
                        .clamp(shouldClamp);
                    // .nice(undefined) still modifies the scale boundaries, and for datetime this messes things up.
                    // we use millisecond ticks since epoch for datetime, so we don"t want any "nice" with numbers like 17398203392.
                    if (niceCount) {
                        scale.nice(niceCount);
                    }
                    return scale;
                }
                HistogramAxisHelper.createLinearScale = createLinearScale;
                function getRecommendedNumberOfTicksForXAxis(availableWidth) {
                    if (availableWidth < 300) {
                        return 3;
                    }
                    if (availableWidth < 500) {
                        return 5;
                    }
                    return 8;
                }
                HistogramAxisHelper.getRecommendedNumberOfTicksForXAxis = getRecommendedNumberOfTicksForXAxis;
                function getRecommendedNumberOfTicksForYAxis(availableWidth) {
                    // if (availableWidth < 40) {
                    //     return 0;
                    // }
                    // if (availableWidth < 75) {
                    //     return 1;
                    // }
                    // if (availableWidth < 100) {
                    //     return 2;
                    // }
                    if (availableWidth < 150) {
                        return 3;
                    }
                    if (availableWidth < 300) {
                        return 5;
                    }
                    return 8;
                }
                HistogramAxisHelper.getRecommendedNumberOfTicksForYAxis = getRecommendedNumberOfTicksForYAxis;
                function isOrdinal(type) {
                    return !!(type && (type.text || type.bool || (type.misc && type.misc.barcode) || (type.geography && type.geography.postalCode)));
                }
                HistogramAxisHelper.isOrdinal = isOrdinal;
                /**
                 * Get the best number of ticks based on minimum value, maximum value,
                 * measure metadata and max tick count.
                 *
                 * @param min The minimum of the data domain.
                 * @param max The maximum of the data domain.
                 * @param valuesMetadata The measure metadata array.
                 * @param maxTickCount The max count of intervals.
                 * @param isDateTime - flag to show single tick when min is equal to max.
                 */
                function getBestNumberOfTicks(min, max, valuesMetadata, maxTickCount, isDateTime) {
                    debug.assert(maxTickCount >= 0, "maxTickCount must be greater or equal to zero");
                    if (isNaN(min) || isNaN(max))
                        return DefaultBestTickCount;
                    debug.assert(min <= max, "min value needs to be less or equal to max value");
                    if (maxTickCount <= 1 || (max <= 1 && min >= -1))
                        return maxTickCount;
                    if (min === max) {
                        // datetime needs to only show one tick value in this case so formatting works correctly
                        if (!!isDateTime)
                            return 1;
                        return DefaultBestTickCount;
                    }
                    if (hasNonIntegerData(valuesMetadata))
                        return maxTickCount;
                    // e.g. 5 - 2 + 1 = 4, => [2,3,4,5]
                    return Math.min(max - min + 1, maxTickCount);
                }
                HistogramAxisHelper.getBestNumberOfTicks = getBestNumberOfTicks;
                function ensureValuesInRange(values, min, max) {
                    debug.assert(min <= max, "min must be less or equal to max");
                    var filteredValues = values.filter(function (v) { return v >= min && v <= max; });
                    if (filteredValues.length < 2)
                        filteredValues = [min, max];
                    return filteredValues;
                }
                HistogramAxisHelper.ensureValuesInRange = ensureValuesInRange;
                function hasNonIntegerData(valuesMetadata) {
                    for (var i = 0, len = valuesMetadata.length; i < len; i++) {
                        var currentMetadata = valuesMetadata[i];
                        if (currentMetadata && currentMetadata.type && !currentMetadata.type.integer) {
                            return true;
                        }
                    }
                    return false;
                }
                HistogramAxisHelper.hasNonIntegerData = hasNonIntegerData;
                /**
                 * Round out very small zero tick values (e.g. -1e-33 becomes 0).
                 *
                 * @param ticks Array of numbers (from d3.scale.ticks([maxTicks])).
                 * @param epsilon Max ratio of calculated tick interval which we will recognize as zero.
                 *
                 * e.g.
                 *     ticks = [-2, -1, 1e-10, 3, 4]; epsilon = 1e-5;
                 *     closeZero = 1e-5 * | 2 - 1 | = 1e-5
                 *     // Tick values <= 1e-5 replaced with 0
                 *     return [-2, -1, 0, 3, 4];
                 */
                function createTrueZeroTickLabel(ticks, epsilon) {
                    if (epsilon === void 0) { epsilon = 1e-5; }
                    if (!ticks || ticks.length < 2)
                        return ticks;
                    var closeZero = epsilon * Math.abs(ticks[1] - ticks[0]);
                    return ticks.map(function (tick) { return Math.abs(tick) <= closeZero ? 0 : tick; });
                }
            })(HistogramAxisHelper = Histogram1445664487616.HistogramAxisHelper || (Histogram1445664487616.HistogramAxisHelper = {}));
            var HistogramBehavior = (function () {
                function HistogramBehavior() {
                }
                HistogramBehavior.create = function () {
                    return new HistogramBehavior();
                };
                HistogramBehavior.prototype.bindEvents = function (behaviorOptions, selectionHandler) {
                    var _this = this;
                    this.columns = behaviorOptions.columns;
                    this.interactivityService = behaviorOptions.interactivityService;
                    this.clearCatcher = behaviorOptions.clearCatcher;
                    this.columns.on("click", function (dataPoint) {
                        selectionHandler.handleClearSelection();
                        if (!HistogramBehavior.areDataPointsSelected(_this.selectedDataPoints, dataPoint.subDataPoints)) {
                            dataPoint.subDataPoints.forEach(function (subDataPoint) {
                                selectionHandler.handleSelection(subDataPoint, true);
                            });
                            _this.selectedDataPoints = dataPoint.subDataPoints;
                        }
                        else {
                            _this.createAnEmptySelectedDataPoints();
                        }
                    });
                    this.clearCatcher.on("click", function () {
                        selectionHandler.handleClearSelection();
                        _this.createAnEmptySelectedDataPoints();
                    });
                };
                HistogramBehavior.prototype.renderSelection = function (hasSelection) {
                    histogramUtils.updateFillOpacity(this.columns, this.interactivityService, hasSelection);
                };
                HistogramBehavior.areDataPointsSelected = function (selectedDataPoints, dataPoints) {
                    if (!dataPoints
                        || !selectedDataPoints
                        || dataPoints.length !== selectedDataPoints.length) {
                        return false;
                    }
                    return selectedDataPoints.every(function (selectedDataPoint) {
                        return dataPoints.some(function (dataPoint) {
                            return selectedDataPoint
                                && dataPoint
                                && selectedDataPoint.identity
                                && dataPoint.identity
                                && selectedDataPoint.identity.equals(dataPoint.identity);
                        });
                    });
                };
                HistogramBehavior.prototype.createAnEmptySelectedDataPoints = function () {
                    this.selectedDataPoints = [];
                };
                return HistogramBehavior;
            }());
            Histogram1445664487616.HistogramBehavior = HistogramBehavior;
            var histogramUtils;
            (function (histogramUtils) {
                histogramUtils.DimmedOpacity = 0.4;
                histogramUtils.DefaultOpacity = 1.0;
                function getFillOpacity(selected, highlight, hasSelection, hasPartialHighlights) {
                    if ((hasPartialHighlights && !highlight) || (hasSelection && !selected)) {
                        return histogramUtils.DimmedOpacity;
                    }
                    return histogramUtils.DefaultOpacity;
                }
                histogramUtils.getFillOpacity = getFillOpacity;
                function getStateOfDataPoint(dataPoint) {
                    var selected = false, highlight = false;
                    if (dataPoint.subDataPoints && dataPoint.subDataPoints.length > 0) {
                        dataPoint.subDataPoints.forEach(function (subDataPoint) {
                            selected = selected || subDataPoint.selected;
                            highlight = highlight || subDataPoint.highlight;
                        });
                    }
                    return {
                        selected: selected,
                        highlight: highlight
                    };
                }
                histogramUtils.getStateOfDataPoint = getStateOfDataPoint;
                function updateFillOpacity(columns, interactivityService, hasSelection) {
                    if (hasSelection === void 0) { hasSelection = false; }
                    var hasHighlights = false;
                    if (interactivityService) {
                        hasHighlights = interactivityService.hasSelection();
                    }
                    columns.style("fill-opacity", function (dataPoint) {
                        var selectedDataPoint = histogramUtils.getStateOfDataPoint(dataPoint);
                        return histogramUtils.getFillOpacity(selectedDataPoint.selected, selectedDataPoint.highlight, !selectedDataPoint.highlight && hasSelection, !selectedDataPoint.selected && hasHighlights);
                    });
                }
                histogramUtils.updateFillOpacity = updateFillOpacity;
            })(histogramUtils = Histogram1445664487616.histogramUtils || (Histogram1445664487616.histogramUtils = {}));
        })(Histogram1445664487616 = visuals.Histogram1445664487616 || (visuals.Histogram1445664487616 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.Histogram1445664487616 = {
                name: 'Histogram1445664487616',
                class: 'Histogram1445664487616',
                capabilities: powerbi.visuals.Histogram1445664487616.Histogram.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.Histogram1445664487616.Histogram(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
