/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PulseChart1459209850231;
        (function (PulseChart1459209850231) {
            var SelectionManager = visuals.utility.SelectionManager;
            var ValueFormatter = powerbi.visuals.valueFormatter;
            var createClassAndSelector = jsCommon.CssConstants.createClassAndSelector;
            var TextMeasurementService = powerbi.TextMeasurementService;
            var defaultLabelPrecision = visuals.dataLabelUtils.defaultLabelPrecision;
            var defaultLabelColor = visuals.dataLabelUtils.defaultLabelColor;
            var DefaultFontSizeInPt = visuals.dataLabelUtils.DefaultFontSizeInPt;
            var defaultLabelDensity = visuals.dataLabelUtils.defaultLabelDensity;
            function createEnumTypeFromEnum(type) {
                var even = false;
                return powerbi.createEnumType(Object.keys(type)
                    .filter(function (key, i) { return ((!!(i % 2)) === even && type[key] === key && !void (even === !even)) || (!!(i % 2)) !== even; })
                    .map(function (x) { return { value: x, displayName: x }; }));
            }
            PulseChart1459209850231.createEnumTypeFromEnum = createEnumTypeFromEnum;
            (function (PulseChartXAxisDateFormat) {
                //DateAndTime = <any>'Date and time',
                PulseChartXAxisDateFormat[PulseChartXAxisDateFormat["DateOnly"] = 'Date only'] = "DateOnly";
                PulseChartXAxisDateFormat[PulseChartXAxisDateFormat["TimeOnly"] = 'Time only'] = "TimeOnly";
            })(PulseChart1459209850231.PulseChartXAxisDateFormat || (PulseChart1459209850231.PulseChartXAxisDateFormat = {}));
            var PulseChartXAxisDateFormat = PulseChart1459209850231.PulseChartXAxisDateFormat;
            (function (XAxisPosition) {
                XAxisPosition[XAxisPosition["Center"] = 'Center'] = "Center";
                XAxisPosition[XAxisPosition["Bottom"] = 'Bottom'] = "Bottom";
            })(PulseChart1459209850231.XAxisPosition || (PulseChart1459209850231.XAxisPosition = {}));
            var XAxisPosition = PulseChart1459209850231.XAxisPosition;
            (function (RunnerCounterPosition) {
                RunnerCounterPosition[RunnerCounterPosition["TopLeft"] = 'Top Left'] = "TopLeft";
                RunnerCounterPosition[RunnerCounterPosition["TopRight"] = 'Top Right'] = "TopRight";
            })(PulseChart1459209850231.RunnerCounterPosition || (PulseChart1459209850231.RunnerCounterPosition = {}));
            var RunnerCounterPosition = PulseChart1459209850231.RunnerCounterPosition;
            var PulseChart = (function () {
                function PulseChart(options) {
                    if (options) {
                        if (options.svg) {
                            this.svg = options.svg;
                        }
                    }
                    this.margin = PulseChart.DefaultMargin;
                }
                PulseChart.getProperties = function (capabilities) {
                    var result = {};
                    for (var objectKey in capabilities.objects) {
                        result[objectKey] = {};
                        for (var propKey in capabilities.objects[objectKey].properties) {
                            result[objectKey][propKey] = {
                                objectName: objectKey,
                                propertyName: propKey
                            };
                        }
                    }
                    return result;
                };
                PulseChart.GetPopupValueTextProperties = function (text, fontSizeValue) {
                    if (fontSizeValue === void 0) { fontSizeValue = 12; }
                    return {
                        text: text || "",
                        fontFamily: "sans-serif",
                        fontSize: fontSizeValue + "px",
                    };
                };
                PulseChart.GetPopupTitleTextProperties = function (text, fontSizeValue) {
                    if (fontSizeValue === void 0) { fontSizeValue = 12; }
                    return {
                        text: text || "",
                        fontFamily: "sans-serif",
                        fontWeight: "bold",
                        fontSize: fontSizeValue + "px",
                    };
                };
                PulseChart.GetPopupDescriptionTextProperties = function (text, fontSizeValue) {
                    if (fontSizeValue === void 0) { fontSizeValue = 12; }
                    return {
                        text: text || "",
                        fontFamily: "sans-serif",
                        fontSize: fontSizeValue + "px",
                    };
                };
                PulseChart.GetRunnerCounterTextProperties = function (text, fontSizeValue) {
                    if (fontSizeValue === void 0) { fontSizeValue = 12; }
                    return {
                        text: text || "",
                        fontFamily: "sans-serif",
                        fontSize: fontSizeValue + "px",
                    };
                };
                PulseChart.ConvertTextPropertiesToStyle = function (textProperties) {
                    return {
                        'font-family': textProperties.fontFamily,
                        'font-weight': textProperties.fontWeight,
                        'font-size': textProperties.fontSize
                    };
                };
                PulseChart.GetDateTimeFormatString = function (dateFormatType, dateFormat) {
                    switch (dateFormatType) {
                        case PulseChartXAxisDateFormat.DateOnly: return dateFormat;
                        case PulseChartXAxisDateFormat.TimeOnly: return "H:mm";
                        default: return "";
                    }
                    ;
                };
                PulseChart.GetFullWidthOfDateFormat = function (dateFormat, textProperties) {
                    textProperties.text = visuals.valueFormatter.create({ format: dateFormat }).format(new Date(2000, 10, 20, 20, 20, 20));
                    return TextMeasurementService.measureSvgTextWidth(textProperties);
                };
                PulseChart.AddOnTouchClick = function (selection, callback) {
                    var preventDefaultCallback = function (data, index) { d3.event.preventDefault(); callback(data, index); };
                    return selection.on("click", preventDefaultCallback).on("touchstart", preventDefaultCallback);
                };
                PulseChart.getCategoricalColumnOfRole = function (dataView, roleName) {
                    var filterFunc = function (cols) { return cols.filter(function (x) { return x.source && x.source.roles && x.source.roles[roleName]; })[0]; };
                    return filterFunc(dataView.categorical.categories) || filterFunc(dataView.categorical.values);
                };
                PulseChart.converter = function (dataView, colors, interactivityService) {
                    if (!dataView
                        || !dataView.categorical
                        || !dataView.categorical.values
                        || !dataView.categorical.values[0]
                        || !dataView.categorical.values[0].values
                        || !dataView.categorical.categories) {
                        return null;
                    }
                    var columns = _.mapValues(PulseChart.RoleNames, function (x, i) { return PulseChart.getCategoricalColumnOfRole(dataView, i); });
                    var timeStampColumn = columns.Timestamp;
                    if (!timeStampColumn) {
                        return null;
                    }
                    var isScalar = !(timeStampColumn.source && timeStampColumn.source.type && timeStampColumn.source.type.dateTime);
                    var settings = PulseChart.parseSettings(dataView, colors, columns);
                    var categoryValues = timeStampColumn.values;
                    if (!categoryValues || _.isEmpty(dataView.categorical.values) || !columns.Value || _.isEmpty(columns.Value.values)) {
                        return null;
                    }
                    var minValuesValue = Math.min.apply(null, columns.Value.values), maxValuesValue = Math.max.apply(null, columns.Value.values);
                    var minCategoryValue = Math.min.apply(null, categoryValues), maxCategoryValue = Math.max.apply(null, categoryValues);
                    settings.xAxis.dateFormat =
                        (maxCategoryValue - minCategoryValue < (24 * 60 * 60 * 1000)
                            && new Date(maxCategoryValue).getDate() === new Date(minCategoryValue).getDate())
                            ? PulseChartXAxisDateFormat.TimeOnly
                            : PulseChartXAxisDateFormat.DateOnly;
                    settings.xAxis.formatterOptions = {
                        value: isScalar ? minCategoryValue : new Date(minCategoryValue),
                        value2: isScalar ? maxCategoryValue : new Date(maxCategoryValue)
                    };
                    settings.yAxis.formatterOptions = {
                        value: minValuesValue,
                        value2: maxValuesValue,
                        format: ValueFormatter.getFormatString(columns.Value.source, PulseChart.DefaultSettings.formatStringProperty)
                    };
                    if (isScalar) {
                        settings.xAxis.formatterOptions.format = ValueFormatter.getFormatString(timeStampColumn.source, PulseChart.DefaultSettings.formatStringProperty);
                    }
                    else {
                        settings.xAxis.formatterOptions.format = PulseChart.GetDateTimeFormatString(settings.xAxis.dateFormat, timeStampColumn.source.format);
                    }
                    var widthOfXAxisLabel = 70;
                    var widthOfTooltipValueLabel = isScalar ? 60 : PulseChart.GetFullWidthOfDateFormat(timeStampColumn.source.format, PulseChart.GetPopupValueTextProperties()) + 5;
                    var heightOfTooltipDescriptionTextLine = TextMeasurementService.measureSvgTextHeight(PulseChart.GetPopupDescriptionTextProperties("lj", settings.popup.fontSize));
                    var runnerCounterFormatString = columns.RunnerCounter && visuals.valueFormatter.getFormatString(columns.RunnerCounter.source, settings.formatStringProperty);
                    settings.popup.width = Math.max(widthOfTooltipValueLabel + 20, settings.popup.width);
                    var minSize = PulseChart.DefaultSettings.dots.minSize;
                    var maxSize = PulseChart.DefaultSettings.dots.maxSize;
                    if (settings.dots) {
                        minSize = settings.dots.minSize;
                        maxSize = settings.dots.maxSize;
                    }
                    var eventSizeScale = PulseChart.createScale(true, columns.EventSize ? [d3.min(columns.EventSize.values), d3.max(columns.EventSize.values)] : [0, 0], minSize, maxSize);
                    var xAxisCardProperties = PulseChartAxisPropertiesHelper.getCategoryAxisProperties(dataView.metadata);
                    var hasDynamicSeries = !!(timeStampColumn.values && timeStampColumn.source);
                    var dataPointLabelSettings = PulseChartDataLabelUtils.getDefaultPulseChartLabelSettings();
                    var gapWidths = PulseChart.getGapWidths(categoryValues);
                    var maxGapWidth = Math.max.apply(null, gapWidths);
                    var firstValueMeasureIndex = 0, firstGroupIndex = 0, secondGroupIndex = 1;
                    var grouped = dataView.categorical.values && dataView.categorical.values.grouped();
                    var y_group0Values = grouped[firstGroupIndex]
                        && grouped[firstGroupIndex].values[firstValueMeasureIndex]
                        && grouped[firstGroupIndex].values[firstValueMeasureIndex].values;
                    var y_group1Values = grouped[secondGroupIndex]
                        && grouped[secondGroupIndex].values[firstValueMeasureIndex]
                        && grouped[secondGroupIndex].values[firstValueMeasureIndex].values;
                    var series = [];
                    var dataPoints = [];
                    for (var categoryIndex = 0, seriesCategoryIndex = 0, len = timeStampColumn.values.length; categoryIndex < len; categoryIndex++, seriesCategoryIndex++) {
                        var categoryValue = categoryValues[categoryIndex];
                        var value = visuals.AxisHelper.normalizeNonFiniteNumber(timeStampColumn.values[categoryIndex]);
                        var runnerCounterValue = columns.RunnerCounter && columns.RunnerCounter.values && columns.RunnerCounter.values[categoryIndex];
                        var identity = visuals.SelectionIdBuilder.builder().withCategory(timeStampColumn, categoryIndex).createSelectionId();
                        var minGapWidth = Math.max((maxCategoryValue - minCategoryValue) / PulseChart.MaxGapCount, PulseChart.MinGapWidth[settings.xAxis.dateFormat]);
                        var gapWidth = gapWidths[categoryIndex];
                        var isGap = settings.gaps.show
                            && gapWidth > 0
                            && gapWidth > (minGapWidth + (100 - settings.gaps.visibleGapsPercentage) * (maxGapWidth - minGapWidth) / 100);
                        if (isGap && dataPoints.length > 0) {
                            series.push({
                                displayName: grouped[firstGroupIndex].name,
                                key: identity.getKey(),
                                lineIndex: series.length,
                                color: settings.series.fill,
                                xCol: timeStampColumn.source,
                                yCol: timeStampColumn.source,
                                data: dataPoints,
                                identity: identity,
                                selected: false,
                                labelSettings: dataPointLabelSettings,
                                width: settings.series.width,
                                widthOfGap: gapWidth
                            });
                            seriesCategoryIndex = 0;
                            dataPoints = [];
                        }
                        // When Scalar, skip null categories and null values so we draw connected lines and never draw isolated dots.
                        if (isScalar && (categoryValue === null || value === null)) {
                            continue;
                        }
                        var popupInfo = null;
                        var eventSize = (columns.EventSize && columns.EventSize.values && columns.EventSize.values[categoryIndex]) || 0;
                        if ((columns.EventTitle && columns.EventTitle.values && columns.EventTitle.values[categoryIndex]) ||
                            (columns.EventDescription && columns.EventDescription.values && columns.EventDescription.values[categoryIndex])) {
                            var formattedValue = categoryValue;
                            if (!isScalar && categoryValue) {
                                formattedValue = visuals.valueFormatter.create({ format: timeStampColumn.source.format }).format(categoryValue);
                            }
                            popupInfo = {
                                value: formattedValue,
                                title: columns.EventTitle && columns.EventTitle.values && columns.EventTitle.values[categoryIndex],
                                description: columns.EventDescription && columns.EventDescription.values && columns.EventDescription.values[categoryIndex],
                            };
                        }
                        var dataPoint = {
                            categoryValue: categoryValue,
                            value: value,
                            categoryIndex: categoryIndex,
                            seriesIndex: series.length,
                            tooltipInfo: null,
                            popupInfo: popupInfo,
                            selected: false,
                            identity: identity,
                            key: JSON.stringify({ ser: identity.getKey(), catIdx: categoryIndex }),
                            labelFill: dataPointLabelSettings.labelColor,
                            labelSettings: dataPointLabelSettings,
                            x: categoryValue,
                            y: (y_group0Values && y_group0Values[categoryIndex]) || (y_group1Values && y_group1Values[categoryIndex]) || 0,
                            pointColor: settings.series.fill,
                            groupIndex: PulseChart.getGroupIndex(categoryIndex, grouped),
                            eventSize: columns.EventSize ? eventSizeScale(eventSize) : 0,
                            runnerCounterValue: runnerCounterValue,
                            runnerCounterFormatString: runnerCounterFormatString,
                            specificIdentity: undefined,
                        };
                        dataPoints.push(dataPoint);
                    }
                    if (interactivityService) {
                        interactivityService.applySelectionStateToData(dataPoints);
                    }
                    if (dataPoints.length > 0) {
                        series.push({
                            displayName: grouped[firstGroupIndex].name,
                            key: identity.getKey(),
                            lineIndex: series.length,
                            color: settings.series.fill,
                            xCol: timeStampColumn.source,
                            yCol: timeStampColumn.source,
                            data: dataPoints,
                            identity: identity,
                            selected: false,
                            labelSettings: dataPointLabelSettings,
                            width: settings.series.width,
                            widthOfGap: 0
                        });
                    }
                    xAxisCardProperties = PulseChartAxisPropertiesHelper.getCategoryAxisProperties(dataView.metadata);
                    var valueAxisProperties = PulseChartAxisPropertiesHelper.getValueAxisProperties(dataView.metadata);
                    var values = dataView.categorical.categories;
                    // Convert to DataViewMetadataColumn
                    var valuesMetadataArray = [];
                    if (values) {
                        for (var i = 0; i < values.length; i++) {
                            if (values[i] && values[i].source && values[i].source.displayName) {
                                valuesMetadataArray.push({ displayName: values[i].source.displayName });
                            }
                        }
                    }
                    var axesLabels = visuals.converterHelper.createAxesLabels(xAxisCardProperties, valueAxisProperties, timeStampColumn.source, valuesMetadataArray);
                    if (interactivityService) {
                        interactivityService.applySelectionStateToData(series);
                    }
                    return {
                        columns: columns,
                        series: series,
                        isScalar: isScalar,
                        dataLabelsSettings: dataPointLabelSettings,
                        axesLabels: { x: axesLabels.xAxisLabel, y: axesLabels.yAxisLabel },
                        hasDynamicSeries: hasDynamicSeries,
                        categoryMetadata: timeStampColumn.source,
                        categories: categoryValues,
                        settings: settings,
                        grouped: grouped,
                        hasHighlights: !!columns.Value.highlights,
                        widthOfXAxisLabel: widthOfXAxisLabel,
                        widthOfTooltipValueLabel: widthOfTooltipValueLabel,
                        heightOfTooltipDescriptionTextLine: heightOfTooltipDescriptionTextLine,
                        runnerCounterHeight: TextMeasurementService.measureSvgTextHeight(PulseChart.GetRunnerCounterTextProperties("lj", settings.runnerCounter.fontSize))
                    };
                };
                PulseChart.createAxisY = function (commonYScale, height, formatterOptions, show) {
                    if (show === void 0) { show = true; }
                    var formatter = visuals.valueFormatter.create(formatterOptions);
                    var ticks = Math.max(2, Math.round(height / 40));
                    var yAxis = d3.svg.axis()
                        .scale(commonYScale)
                        .ticks(ticks)
                        .outerTickSize(0)
                        .tickFormat(formatter.format);
                    return yAxis;
                };
                PulseChart.createAxisX = function (isScalar, series, originalScale, formatterOptions, dateFormat, position, widthOfXAxisLabel) {
                    var scales = PulseChart.getXAxisScales(series, isScalar, originalScale);
                    var xAxisProperties = new Array(scales.length);
                    for (var i = 0, rotate = false; i < xAxisProperties.length; i++) {
                        var values = PulseChart.getXAxisValuesToDisplay(scales[i], rotate, isScalar, dateFormat, widthOfXAxisLabel);
                        if (!rotate
                            && position === XAxisPosition.Bottom
                            && values.length < PulseChart.MinimumTicksToRotate) {
                            var rotatedValues = PulseChart.getXAxisValuesToDisplay(scales[i], true, isScalar, dateFormat, widthOfXAxisLabel);
                            if (rotatedValues.length > values.length) {
                                rotate = true;
                                i = -1;
                                continue;
                            }
                        }
                        xAxisProperties[i] = { values: values, scale: scales[i], rotate: rotate };
                    }
                    formatterOptions.tickCount = xAxisProperties.length && xAxisProperties.map(function (x) { return x.values.length; }).reduce(function (a, b) { return a + b; }) * 5;
                    formatterOptions.value = originalScale.domain()[0];
                    formatterOptions.value2 = originalScale.domain()[1];
                    xAxisProperties.forEach(function (properties) {
                        var values = properties.values.filter(function (value) { return value !== null; });
                        var formatter = visuals.valueFormatter.create(formatterOptions);
                        properties.axis = d3.svg.axis()
                            .scale(properties.scale)
                            .tickValues(values)
                            .tickFormat(formatter.format)
                            .outerTickSize(0);
                    });
                    return xAxisProperties;
                };
                PulseChart.getXAxisScales = function (series, isScalar, originalScale) {
                    return series.map(function (seriesElement) {
                        var dataPoints = seriesElement.data, minValue = dataPoints[0].categoryValue, maxValue = dataPoints[dataPoints.length - 1].categoryValue, minX = originalScale(dataPoints[0].categoryValue), maxX = originalScale(dataPoints[dataPoints.length - 1].categoryValue);
                        return PulseChart.createScale(isScalar, [minValue, maxValue], minX, maxX);
                    });
                };
                PulseChart.getXAxisValuesToDisplay = function (scale, rotate, isScalar, dateFormat, widthOfXAxisLabel) {
                    var genScale = scale;
                    var tickWidth = rotate
                        ? PulseChart.XAxisTickHeight * (rotate ? Math.abs(Math.sin(PulseChart.AxisTickRotateAngle * Math.PI / 180)) : 0)
                        : widthOfXAxisLabel;
                    var tickSpace = PulseChart.XAxisTickSpace;
                    if (scale.range()[1] < tickWidth) {
                        return [];
                    }
                    var minValue = scale.invert(scale.range()[0] + tickWidth / 2);
                    var maxValue = scale.invert(scale.range()[1] - tickWidth / 2);
                    var width = scale.range()[1] - scale.range()[0];
                    var maxTicks = Math.floor((width + tickSpace) / (tickWidth + tickSpace));
                    if (rotate) {
                        maxTicks = Math.min(PulseChart.MinimumTicksToRotate, maxTicks);
                    }
                    var values = [];
                    if (isScalar) {
                        values = d3.range(minValue, maxValue, (maxValue - minValue) / (maxTicks * 100));
                    }
                    else {
                        values = (dateFormat === PulseChartXAxisDateFormat.TimeOnly ? d3.time.minute : d3.time.day)
                            .range(minValue, maxValue);
                    }
                    if (!values.length || _.last(values) < maxValue) {
                        values.push(maxValue);
                    }
                    if (!maxTicks) {
                        return [];
                    }
                    maxTicks = Math.min(values.length, maxTicks);
                    var valuesIndexses = d3.scale.ordinal().domain(d3.range(maxTicks)).rangePoints([0, values.length - 1]).range(); //randeRoundPoints is not defined
                    values = valuesIndexses.map(function (x) { return values[Math.round(x)]; });
                    for (var i = 1; i < values.length; i++) {
                        var prevXValue = genScale(values[i - 1]);
                        var curXValue = genScale(values[i]);
                        if (curXValue - prevXValue < tickWidth + tickSpace / 3) {
                            values.splice(i--, 1);
                        }
                    }
                    return values;
                };
                PulseChart.getGroupIndex = function (index, grouped) {
                    for (var i = 0; i < grouped.length; i++) {
                        if (grouped[i].values && grouped[i].values[0] &&
                            grouped[i].values[0].values[index] !== undefined &&
                            grouped[i].values[0].values[index] !== null) {
                            return i;
                        }
                    }
                    return 0;
                };
                PulseChart.getGapWidths = function (values) {
                    var result = [];
                    for (var i = 0, prevVal = 0, length = values.length; i < length; i++) {
                        if (!prevVal || !values[i]) {
                            result.push(0);
                        }
                        else {
                            result.push(values[i] - prevVal);
                        }
                        prevVal = values[i];
                    }
                    return result;
                };
                PulseChart.createScale = function (isScalar, domain, minX, maxX) {
                    var scale;
                    if (isScalar) {
                        scale = d3.scale.linear();
                    }
                    else {
                        scale = d3.time.scale();
                    }
                    return scale
                        .domain(domain)
                        .range([minX, maxX]);
                };
                Object.defineProperty(PulseChart.prototype, "runnerCounterPlaybackButtonsHeight", {
                    get: function () {
                        return Math.max(PulseChart.PlaybackButtonsHeight, this.data && (this.data.runnerCounterHeight / 2 + 17));
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseChart.prototype, "popupHeight", {
                    get: function () {
                        return this.data
                            && this.data.settings
                            && this.data.settings.popup
                            && this.data.settings.popup.show
                            && this.data.settings.popup.height || 0;
                    },
                    enumerable: true,
                    configurable: true
                });
                PulseChart.prototype.init = function (options) {
                    powerbi.formattingService.initialize(); //Fixes the framework bug: "Cannot read property 'getFormatString' of undefined".
                    this.host = options.host;
                    this.selectionManager = new SelectionManager({ hostServices: this.host });
                    var svg = this.svg = d3.select(options.element.get(0))
                        .append('svg')
                        .classed('pulseChart', true);
                    this.gaps = svg.append('g').classed(PulseChart.Gaps.class, true);
                    this.yAxis = svg.append('g').attr('class', 'y axis');
                    this.chart = svg.append('g').attr('class', PulseChart.Chart.class);
                    this.dots = svg.append('g').attr('class', 'dots');
                    this.animationDot = this.dots.append('circle').classed(PulseChart.AnimationDot.class, true).attr('display', 'none');
                    this.animationHandler = new PulseAnimator(this, svg);
                    var style = options.style;
                    this.colors = style && style.colorPalette
                        ? style.colorPalette.dataColors
                        : new visuals.DataColorPalette();
                };
                PulseChart.prototype.update = function (options) {
                    if (!options || !options.dataViews || !options.dataViews[0]) {
                        return;
                    }
                    this.viewport = $.extend({}, options.viewport);
                    var dataView = options.dataViews[0];
                    this.updateData(PulseChart.converter(dataView, this.colors));
                    if (!this.validateData(this.data)) {
                        this.clearAll(true);
                        return;
                    }
                    var width = this.getChartWidth();
                    this.calculateXAxisProperties(width);
                    if (this.data.xScale.ticks(undefined).length < 2) {
                        this.clearAll(true);
                        return;
                    }
                    var height = this.getChartHeight(this.data.settings.xAxis.show
                        && this.data.series.some(function (series) { return series.xAxisProperties.rotate; }));
                    this.calculateYAxisProperties(height);
                    this.size = { width: width, height: height };
                    this.updateElements();
                    this.render(true);
                };
                PulseChart.prototype.updateData = function (data) {
                    if (!this.data) {
                        this.data = data;
                        return;
                    }
                    var oldDataObj = this.getDataArrayToCompare(this.data);
                    var newDataObj = this.getDataArrayToCompare(data);
                    if (!_.isEqual(oldDataObj, newDataObj)) {
                        this.clearAll(false);
                    }
                    this.data = data;
                };
                PulseChart.prototype.getDataArrayToCompare = function (data) {
                    if (!data || !data.series) {
                        return null;
                    }
                    var dataPoints = _.flatten(data.series.map(function (x) { return x.data; }));
                    return _.flatten(dataPoints.map(function (x) {
                        return x && _.flatten([
                            [
                                x.categoryValue,
                                x.eventSize,
                                x.groupIndex,
                                x.runnerCounterValue,
                                x.y,
                                x.seriesIndex
                            ],
                            x.popupInfo && [x.popupInfo.description, x.popupInfo.title, x.popupInfo.value]
                        ]);
                    }));
                };
                PulseChart.prototype.validateData = function (data) {
                    if (!data) {
                        return false;
                    }
                    if (data.categories.some(function (x) { return !(x instanceof Date || $.isNumeric(x)); })) {
                        return false;
                    }
                    return true;
                };
                PulseChart.prototype.getChartWidth = function () {
                    var marginRight = this.margin.right;
                    if (this.data.settings.yAxis && this.data.settings.yAxis.show) {
                        marginRight += PulseChart.MaxWidthOfYAxis;
                    }
                    var width = this.viewport.width - this.margin.left - marginRight;
                    return Math.max(width, PulseChart.DefaultViewport.width);
                };
                PulseChart.prototype.getChartHeight = function (xAxisRotated) {
                    var marginBottom = 10 + (xAxisRotated
                        ? this.data.widthOfXAxisLabel * Math.abs(Math.sin(PulseChart.AxisTickRotateAngle * Math.PI / 180))
                        : 3);
                    if (!this.data.settings.popup.alwaysOnTop && this.popupHeight) {
                        marginBottom = Math.max(this.margin.bottom + this.popupHeight, marginBottom);
                    }
                    var height = this.viewport.height
                        - this.margin.top
                        - this.runnerCounterPlaybackButtonsHeight
                        - marginBottom
                        - this.popupHeight;
                    return Math.max(height, PulseChart.DefaultViewport.height);
                };
                PulseChart.prototype.updateElements = function () {
                    var chartMarginTop = this.margin.top + this.runnerCounterPlaybackButtonsHeight + this.popupHeight;
                    this.svg.attr(this.viewport);
                    this.svg.style('display', undefined);
                    this.gaps.attr('transform', visuals.SVGUtil.translate(this.margin.left, chartMarginTop + (this.size.height / 2)));
                    this.chart.attr('transform', visuals.SVGUtil.translate(this.margin.left, chartMarginTop));
                    this.yAxis.attr('transform', visuals.SVGUtil.translate(this.size.width + this.margin.left, chartMarginTop));
                    this.dots.attr('transform', visuals.SVGUtil.translate(this.margin.left, chartMarginTop));
                };
                PulseChart.prototype.calculateXAxisProperties = function (width) {
                    this.data.xScale = PulseChart.createScale(this.data.isScalar, [this.data.categories[0], this.data.categories[this.data.categories.length - 1]], 0, width);
                    var xAxisProperties = PulseChart.createAxisX(this.data.isScalar, this.data.series, this.data.xScale, $.extend({}, this.data.settings.xAxis.formatterOptions), this.data.settings.xAxis.dateFormat, this.data.settings.xAxis.position, this.data.widthOfXAxisLabel);
                    this.data.series.forEach(function (series, index) {
                        series.xAxisProperties = xAxisProperties[index];
                    });
                };
                PulseChart.prototype.calculateYAxisProperties = function (height) {
                    this.data.yScales = this.getYAxisScales(height);
                    var domain = [];
                    this.data.yScales.forEach(function (scale) { return domain = domain.concat(scale.domain()); });
                    this.data.commonYScale = PulseChart.createScale(true, [d3.max(domain), d3.min(domain)], 0, height);
                    this.data.yAxis = PulseChart.createAxisY(this.data.commonYScale, height, this.data.settings.yAxis.formatterOptions);
                };
                PulseChart.prototype.getYAxisScales = function (height) {
                    var data = this.data, stepOfHeight = height / data.grouped.length;
                    return data.grouped.map(function (group, index) {
                        var values = group.values[0].values.map(function (x) { return x || 0; });
                        var minValue = Number.MAX_VALUE, maxValue = -Number.MAX_VALUE;
                        values.forEach(function (value) {
                            if (value < minValue) {
                                minValue = value;
                            }
                            if (value > maxValue) {
                                maxValue = value;
                            }
                        });
                        if (maxValue === minValue) {
                            var offset = maxValue === 0 ? 1 : Math.abs(maxValue / 2);
                            maxValue += offset;
                            minValue -= offset;
                        }
                        return PulseChart.createScale(true, [maxValue, minValue], stepOfHeight * index, stepOfHeight * (index + 1));
                    });
                };
                Object.defineProperty(PulseChart.prototype, "autoplayPauseDuration", {
                    get: function () {
                        return 1000 * ((this.data && this.data.settings && this.data.settings.playback)
                            ? this.data.settings.playback.autoplayPauseDuration
                            : PulseChart.DefaultSettings.playback.autoplayPauseDuration);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseChart.prototype, "isAutoPlay", {
                    get: function () {
                        return this.data &&
                            this.data.settings &&
                            this.data.settings.playback &&
                            this.data.settings.playback.autoplay;
                    },
                    enumerable: true,
                    configurable: true
                });
                PulseChart.prototype.render = function (suppressAnimations) {
                    var duration = visuals.AnimatorCommon.GetAnimationDuration(this.animator, suppressAnimations);
                    var data = this.data;
                    this.lastSelectedPoint = null;
                    var xScale = data.xScale, yScales = data.yScales;
                    this.lineX = d3.svg.line()
                        .x(function (d) {
                        return xScale(d.categoryValue);
                    })
                        .y(function (d) {
                        return yScales[d.groupIndex](d.y);
                    });
                    if (this.data &&
                        this.data.settings &&
                        this.data.settings.playback &&
                        this.data.settings.playback.color) {
                        this.animationHandler.setControlsColor(this.data.settings.playback.color);
                    }
                    this.animationHandler.render();
                    this.animationHandler.setRunnerCounterValue();
                    this.renderAxes(data, duration);
                    this.renderGaps(data, duration);
                };
                PulseChart.prototype.renderAxes = function (data, duration) {
                    this.renderXAxis(data, duration);
                    this.renderYAxis(data, duration);
                };
                PulseChart.prototype.renderXAxis = function (data, duration) {
                    var axisNodeSelection, axisNodeUpdateSelection, 
                    //ticksSelection: D3.Selection,
                    axisBoxUpdateSelection, color = PulseChart.DefaultSettings.xAxis.color, fontColor = PulseChart.DefaultSettings.xAxis.fontColor;
                    if (this.data && this.data.settings && this.data.settings.xAxis) {
                        color = this.data.settings.xAxis.color;
                        fontColor = this.data.settings.xAxis.fontColor;
                    }
                    axisNodeSelection = this.rootSelection.selectAll(PulseChart.XAxisNode.selector);
                    axisNodeUpdateSelection = axisNodeSelection.data(data.series);
                    axisNodeUpdateSelection
                        .enter()
                        .insert("g", "g." + PulseChart.LineContainer.class)
                        .classed(PulseChart.XAxisNode.class, true);
                    axisNodeUpdateSelection
                        .call(function (selection) {
                        selection.forEach(function (selectionElement, index) {
                            d3.select(selectionElement[0])
                                .call(data.series[index].xAxisProperties.axis.orient('bottom'));
                        });
                    });
                    axisNodeUpdateSelection
                        .exit()
                        .remove();
                    axisBoxUpdateSelection = axisNodeUpdateSelection
                        .selectAll(".tick")
                        .selectAll(".axisBox")
                        .data([[]]);
                    axisBoxUpdateSelection
                        .enter()
                        .insert("rect", "text")
                        .classed("axisBox", true);
                    axisBoxUpdateSelection
                        .style('display', this.data.settings.xAxis.position === XAxisPosition.Center ? 'inherit' : 'none')
                        .style('fill', this.data.settings.xAxis.backgroundColor);
                    var tickRectY = this.data.settings.xAxis.position === XAxisPosition.Center ? -11 : 0;
                    axisBoxUpdateSelection.attr({
                        x: -(this.data.widthOfXAxisLabel / 2),
                        y: tickRectY + "px",
                        width: this.data.widthOfXAxisLabel,
                        height: PulseChart.XAxisTickHeight + "px"
                    });
                    axisBoxUpdateSelection
                        .exit()
                        .remove();
                    axisNodeUpdateSelection
                        .style('stroke', this.data.settings.xAxis.position === XAxisPosition.Center ? color : "none")
                        .style('display', this.data.settings.xAxis.show ? 'inherit' : 'none');
                    axisNodeUpdateSelection.call(function (selection) {
                        var rotate = selection.datum().xAxisProperties.rotate;
                        var rotateCoeff = rotate ? Math.abs(Math.sin(PulseChart.AxisTickRotateAngle * Math.PI / 180)) : 0;
                        var dy = tickRectY + 3;
                        selection.selectAll("text")
                            .attr('transform', function () {
                            return "translate(0, " + (dy + 9 + ($(this).width() / 2) * rotateCoeff) + ") rotate(" + (rotate ? PulseChart.AxisTickRotateAngle : 0) + ")";
                        })
                            .style('fill', fontColor)
                            .style('stroke', "none")
                            .attr('dy', -9);
                    });
                    axisNodeUpdateSelection.selectAll(".domain")
                        .style('stroke', color);
                    axisNodeUpdateSelection.selectAll(".domain").forEach(function (element) {
                        $(element).insertBefore($(element).parent().children().first());
                    });
                    var xAxisTop = this.size.height;
                    switch (this.data.settings.xAxis.position) {
                        case XAxisPosition.Center:
                            xAxisTop = xAxisTop / 2;
                            break;
                        case XAxisPosition.Bottom:
                            break;
                    }
                    axisNodeUpdateSelection.attr('transform', visuals.SVGUtil.translate(0, xAxisTop));
                };
                PulseChart.prototype.renderYAxis = function (data, duration) {
                    var yAxis = data.yAxis, isShow = false, color = PulseChart.DefaultSettings.yAxis.color, fontColor = PulseChart.DefaultSettings.yAxis.fontColor;
                    ;
                    yAxis.orient('right');
                    if (this.data &&
                        this.data.settings &&
                        this.data.settings.yAxis &&
                        this.data.settings.yAxis.show) {
                        isShow = true;
                    }
                    if (this.data &&
                        this.data.settings &&
                        this.data.settings.yAxis &&
                        this.data.settings.yAxis) {
                        color = this.data.settings.yAxis.color;
                        fontColor = this.data.settings.yAxis.fontColor;
                    }
                    this.yAxis
                        .call(yAxis)
                        .attr('display', isShow ? 'inline' : 'none');
                    this.yAxis.selectAll('.domain, path, line').style('stroke', color);
                    this.yAxis.selectAll('text').style('fill', fontColor);
                    this.yAxis.selectAll('g.tick line')
                        .attr('x1', -this.size.width);
                };
                PulseChart.prototype.renderChart = function () {
                    var data = this.data;
                    var series = this.data.series;
                    var selection = this.rootSelection = this.chart.selectAll(PulseChart.LineNode.selector).data(series);
                    var lineNode = selection
                        .enter()
                        .append('g')
                        .classed(PulseChart.LineNode.class, true);
                    lineNode
                        .append('g')
                        .classed(PulseChart.LineContainer.class, true);
                    lineNode
                        .append('g')
                        .classed(PulseChart.TooltipContainer.class, true);
                    lineNode
                        .append('g')
                        .classed(PulseChart.DotsContainer.class, true);
                    if (this.animationHandler.isAnimated) {
                        this.showAnimationDot();
                    }
                    else {
                        this.hideAnimationDot();
                    }
                    this.drawTooltips(data, this.selectionManager.getSelectionIds());
                    this.drawDots(data);
                    this.drawLines(data);
                    selection
                        .exit()
                        .remove();
                };
                PulseChart.prototype.drawLinesStatic = function (limit, isAnimated) {
                    var _this = this;
                    var node = PulseChart.Line, nodeParent = PulseChart.LineContainer, rootSelection = this.rootSelection;
                    var selection = rootSelection
                        .filter(function (d, index) { return !isAnimated || index < limit; })
                        .select(nodeParent.selector)
                        .selectAll(node.selector).data(function (d) { return [d]; });
                    selection
                        .enter()
                        .append('path')
                        .classed(node.class, true);
                    selection
                        .style({
                        'fill': "none",
                        'stroke': function (d) { return d.color; },
                        'stroke-width': function (d) { return (d.width + "px"); }
                    });
                    selection.attr('d', function (d) { return _this.lineX(d.data); });
                    selection
                        .exit()
                        .remove();
                };
                PulseChart.prototype.drawLinesStaticBeforeAnimation = function (limit) {
                    var _this = this;
                    var node = PulseChart.Line, nodeParent = PulseChart.LineContainer, rootSelection = this.rootSelection;
                    this.animationSelection = rootSelection.filter(function (d, index) {
                        return index === limit;
                    }).select(nodeParent.selector).selectAll(node.selector).data(function (d) { return [d]; });
                    this.animationSelection
                        .enter()
                        .append('path')
                        .classed(node.class, true);
                    this.animationSelection
                        .style({
                        'fill': "none",
                        'stroke': function (d) { return d.color; },
                        'stroke-width': function (d) { return (d.width + "px"); }
                    });
                    this.animationSelection
                        .attr('d', function (d) {
                        var flooredStart = _this.animationHandler.flooredPosition.index;
                        if (flooredStart === 0) {
                            _this.moveAnimationDot(d.data[0]);
                            return _this.lineX([]);
                        }
                        else {
                            var dataReduced = d.data.slice(0, flooredStart + 1);
                            _this.moveAnimationDot(dataReduced[dataReduced.length - 1]);
                            return _this.lineX(dataReduced);
                        }
                    });
                    this.animationSelection
                        .exit()
                        .remove();
                };
                PulseChart.prototype.moveAnimationDot = function (d) {
                    var xScale = this.data.xScale, yScales = this.data.yScales;
                    this.animationDot
                        .attr("cx", xScale(d.x))
                        .attr("cy", yScales[d.groupIndex](d.y));
                };
                PulseChart.prototype.playAnimation = function (delay) {
                    var _this = this;
                    if (delay === void 0) { delay = 0; }
                    var flooredStart = this.animationHandler.flooredPosition.index;
                    this.showAnimationDot();
                    this.animationSelection
                        .transition()
                        .delay(delay)
                        .duration(this.animationDuration)
                        .ease("linear")
                        .attrTween('d', function (d, index) { return _this.getInterpolation(d.data, flooredStart); })
                        .each("end", function (series) { return _this.handleSelection(_this.animationHandler.flooredPosition); });
                };
                PulseChart.prototype.pauseAnimation = function () {
                    if (!this.animationSelection) {
                        return;
                    }
                    this.hideAnimationDot();
                    this.animationSelection.selectAll("path").transition();
                    this.animationSelection
                        .transition()
                        .duration(0)
                        .delay(0);
                };
                PulseChart.prototype.stopAnimation = function () {
                    this.pauseAnimation();
                    d3.timer.flush();
                };
                PulseChart.prototype.findNextPoint = function (position) {
                    for (var i = position.series; i < this.data.series.length; i++) {
                        var series = this.data.series[i];
                        for (var j = (i === position.series) ? Math.floor(position.index + 1) : 0; j < series.data.length; j++) {
                            if (series.data[j] && series.data[j].popupInfo) {
                                return {
                                    series: i,
                                    index: j
                                };
                            }
                        }
                    }
                    return null;
                };
                PulseChart.prototype.findPrevPoint = function (position) {
                    for (var i = position.series; i >= 0; i--) {
                        var series = this.data.series[i];
                        for (var j = (i === position.series) ? Math.ceil(position.index - 1) : series.data.length; j >= 0; j--) {
                            if (series.data[j] && series.data[j].popupInfo) {
                                return {
                                    series: i,
                                    index: j
                                };
                            }
                        }
                    }
                    return null;
                };
                PulseChart.prototype.isAnimationSeriesAndIndexLast = function (position) {
                    return this.isAnimationSeriesLast(position) && this.isAnimationIndexLast(position);
                };
                PulseChart.prototype.isAnimationSeriesLast = function (position) {
                    return (position.series >= (this.data.series.length - 1));
                };
                PulseChart.prototype.isAnimationIndexLast = function (position) {
                    var series = this.data.series[position.series];
                    return (position.index >= (series.data.length - 1));
                };
                PulseChart.prototype.drawLines = function (data) {
                    var positionSeries = this.animationHandler.position.series, isAnimated = this.animationHandler.isAnimated;
                    this.drawLinesStatic(positionSeries, isAnimated);
                    if (isAnimated) {
                        this.drawLinesStaticBeforeAnimation(positionSeries);
                    }
                };
                PulseChart.prototype.showAnimationDot = function () {
                    if (!this.animationHandler.isPlaying) {
                        return;
                    }
                    var size = PulseChart.DefaultSettings.dots.size;
                    if (this.data &&
                        this.data.settings &&
                        this.data.settings.dots &&
                        this.data.settings.dots.size) {
                        size = this.data.settings.dots.size;
                    }
                    this.animationDot
                        .attr('display', 'inline')
                        .attr("fill", this.data.settings.dots.color)
                        .style("opacity", this.dotOpacity)
                        .attr("r", size);
                };
                PulseChart.prototype.hideAnimationDot = function () {
                    this.animationDot.attr('display', 'none');
                };
                PulseChart.prototype.getInterpolation = function (data, start) {
                    var _this = this;
                    if (!this.data) {
                        return;
                    }
                    var xScale = this.data.xScale, yScales = this.data.yScales;
                    var stop = start + 1;
                    this.showAnimationDot();
                    var lineFunction = d3.svg.line()
                        .x(function (d) { return d.x; })
                        .y(function (d) { return d.y; })
                        .interpolate("linear");
                    var interpolatedLine = data.slice(0, start + 1).map(function (d) {
                        return {
                            x: xScale(d.x),
                            y: yScales[d.groupIndex](d.y)
                        };
                    });
                    var x0 = xScale(data[start].x);
                    var x1 = xScale(data[stop].x);
                    var y0 = yScales[data[start].groupIndex](data[start].y);
                    var y1 = yScales[data[stop].groupIndex](data[stop].y);
                    var interpolateIndex = d3.scale.linear()
                        .domain([0, 1])
                        .range([start, stop]);
                    var interpolateX = d3.scale.linear()
                        .domain([0, 1])
                        .range([x0, x1]);
                    var interpolateY = d3.scale.linear()
                        .domain([0, 1])
                        .range([y0, y1]);
                    this.animationHandler.setRunnerCounterValue(start);
                    return function (t) {
                        if (!_this.animationHandler.isPlaying) {
                            return lineFunction(interpolatedLine);
                        }
                        var x = interpolateX(t);
                        var y = interpolateY(t);
                        _this.animationDot
                            .attr("cx", x)
                            .attr("cy", y);
                        interpolatedLine.push({ "x": x, "y": y });
                        _this.animationHandler.position.index = interpolateIndex(t);
                        return lineFunction(interpolatedLine);
                    };
                };
                PulseChart.prototype.clearSelection = function () {
                    if (this.data && this.data.hasHighlights || !this.selectionManager.hasSelection()) {
                        this.selectionManager.selectedIds = [];
                    }
                    else {
                        this.selectionManager.clear();
                    }
                    this.chart.selectAll(PulseChart.Tooltip.selector).remove();
                };
                PulseChart.prototype.handleSelection = function (position) {
                    var _this = this;
                    if (!this.data) {
                        return;
                    }
                    var animationPlayingIndex = this.animationHandler.animationPlayingIndex;
                    var dataPoint = this.data.series[position.series].data[position.index];
                    var isLastDataPoint = this.animationHandler.isPlaying && this.isAnimationSeriesAndIndexLast(position);
                    if ((!dataPoint || !dataPoint.popupInfo) && (this.animationHandler.isPlaying)) {
                        if (isLastDataPoint) {
                            setTimeout(function () { return _this.animationHandler.toEnd(); }, 0);
                        }
                        else {
                            this.animationHandler.play(0, true);
                        }
                        return;
                    }
                    if (isLastDataPoint) {
                        setTimeout(function () { return _this.animationHandler.toEnd(); }, 0);
                    }
                    else {
                        this.animationHandler.pause();
                    }
                    this.selectionManager.select(dataPoint.identity).then(function (selectionIds) {
                        _this.setSelection(selectionIds);
                        clearTimeout(_this.handleSelectionTimeout);
                        _this.handleSelectionTimeout = setTimeout(function () {
                            if (_this.animationHandler.animationPlayingIndex !== animationPlayingIndex) {
                                return;
                            }
                            if (isLastDataPoint || _this.animationHandler.isPaused) {
                                _this.clearSelection();
                            }
                            if (!isLastDataPoint && _this.animationHandler.isPaused) {
                                _this.animationHandler.play();
                            }
                        }, _this.pauseDuration);
                    });
                };
                Object.defineProperty(PulseChart.prototype, "animationDuration", {
                    get: function () {
                        return 1000 / ((this.data && this.data.settings && this.data.settings.playback)
                            ? this.data.settings.playback.playSpeed
                            : PulseChart.DefaultSettings.playback.playSpeed);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseChart.prototype, "pauseDuration", {
                    get: function () {
                        return 1000 * ((this.data && this.data.settings && this.data.settings.playback)
                            ? this.data.settings.playback.pauseDuration
                            : PulseChart.DefaultSettings.playback.pauseDuration);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseChart.prototype, "dotOpacity", {
                    get: function () {
                        return 1 - ((this.data && this.data.settings && this.data.settings.dots)
                            ? this.data.settings.dots.transparency
                            : PulseChart.DefaultSettings.dots.transparency) / 100;
                    },
                    enumerable: true,
                    configurable: true
                });
                PulseChart.prototype.drawDots = function (data) {
                    var _this = this;
                    if (!data || !data.xScale) {
                        return;
                    }
                    var xScale = data.xScale, yScales = data.yScales, node = PulseChart.Dot, nodeParent = PulseChart.DotsContainer, rootSelection = this.rootSelection, dotColor = this.data.settings.dots.color, dotSize = this.data.settings.dots.size, isAnimated = this.animationHandler.isAnimated, position = this.animationHandler.position;
                    var selection = rootSelection.filter(function (d, index) { return !isAnimated || index <= position.series; })
                        .select(nodeParent.selector)
                        .selectAll(node.selector)
                        .data(function (d, seriesIndex) {
                        return _.filter(d.data, function (value, valueIndex) {
                            if (isAnimated && (seriesIndex === position.series) && (valueIndex > position.index)) {
                                return false;
                            }
                            return (!!value.popupInfo);
                        });
                    });
                    selection
                        .enter()
                        .append("circle")
                        .classed(node.class, true);
                    selection
                        .attr("cx", function (d) { return xScale(d.categoryValue); })
                        .attr("cy", function (d) { return yScales[d.groupIndex](d.y); })
                        .attr("r", function (d) { return d.eventSize || dotSize; })
                        .style("fill", dotColor)
                        .style("opacity", this.dotOpacity)
                        .style("cursor", "pointer")
                        .call(PulseChart.AddOnTouchClick, function (d) {
                        d3.event.stopPropagation();
                        _this.selectionManager.select(d.identity, d3.event.ctrlKey)
                            .then(function (selectionIds) { return _this.setSelection(selectionIds); });
                    });
                    selection
                        .exit()
                        .remove();
                };
                PulseChart.prototype.renderGaps = function (data, duration) {
                    var gaps, gapsSelection, gapsEnterSelection, gapNodeSelection, series = data.series, isScalar = data.isScalar, xScale = data.xScale;
                    gaps = [{
                            left: -4.5,
                            top: -5,
                            height: 10,
                            width: 3
                        }, {
                            left: 1.5,
                            top: -5,
                            height: 10,
                            width: 3
                        }];
                    gapsSelection = this.gaps.selectAll(PulseChart.Gap.selector)
                        .data(series.slice(0, series.length - 1));
                    gapsEnterSelection = gapsSelection
                        .enter()
                        .append("g");
                    gapsSelection
                        .attr("transform", function (seriesElement, index) {
                        var x, middleOfGap = seriesElement.widthOfGap / 2, categoryValue = seriesElement.data[seriesElement.data.length - 1].categoryValue;
                        if (isScalar) {
                            x = xScale(middleOfGap + categoryValue);
                        }
                        else {
                            x = xScale(new Date(middleOfGap + (categoryValue.getTime())));
                        }
                        return visuals.SVGUtil.translate(x, 0);
                    });
                    gapNodeSelection = gapsSelection.selectAll(PulseChart.GapNode.selector)
                        .data(gaps);
                    gapNodeSelection
                        .enter()
                        .append("rect")
                        .attr({
                        x: function (gap) { return gap.left; },
                        y: function (gap) { return gap.top; },
                        height: function (gap) { return gap.height; },
                        width: function (gap) { return gap.width; }
                    })
                        .classed(PulseChart.GapNode.class, true);
                    gapsEnterSelection.classed(PulseChart.Gap.class, true);
                    gapsSelection
                        .exit()
                        .remove();
                    gapNodeSelection
                        .exit()
                        .remove();
                };
                PulseChart.prototype.setSelection = function (selectionIds) {
                    if (!this.data) {
                        return;
                    }
                    this.drawDots(this.data);
                    this.drawTooltips(this.data, selectionIds);
                };
                PulseChart.prototype.isPopupShow = function (d, selectionIds) {
                    if (!this.popupHeight || !d.popupInfo) {
                        return false;
                    }
                    if (selectionIds) {
                        return SelectionManager.containsSelection(selectionIds, d.identity);
                    }
                    return false;
                };
                PulseChart.prototype.drawTooltips = function (data, selectionIds) {
                    var _this = this;
                    var xScale = data.xScale, yScales = data.yScales, node = PulseChart.Tooltip, nodeParent = PulseChart.TooltipContainer;
                    var rootSelection = this.rootSelection;
                    var line = d3.svg.line()
                        .x(function (d) { return d.x; })
                        .y(function (d) { return d.y; });
                    var marginTop = PulseChart.DefaultTooltipSettings.marginTop;
                    var width = this.data.settings.popup.width;
                    var height = this.data.settings.popup.height;
                    var topShift = 20;
                    var tooltipShiftY = function (y, groupIndex) { return _this.isHigherMiddle(y, groupIndex) ? (-1 * marginTop + topShift) : _this.size.height + marginTop; };
                    var tooltipRoot = rootSelection.select(nodeParent.selector).selectAll(node.selector)
                        .data(function (d) {
                        return _.filter(d.data, function (value) { return _this.isPopupShow(value, selectionIds); });
                    });
                    tooltipRoot
                        .enter()
                        .append("g")
                        .classed(node.class, true);
                    tooltipRoot
                        .attr("transform", function (d) {
                        var x = xScale(d.x) - width / 2;
                        var y = tooltipShiftY(d.y, d.groupIndex);
                        d.popupInfo.offsetX = Math.min(_this.viewport.width - _this.margin.right - width, Math.max(-_this.margin.left, x)) - x;
                        return visuals.SVGUtil.translate(x + d.popupInfo.offsetX, y);
                    });
                    var tooltipRect = tooltipRoot.selectAll(PulseChart.TooltipRect.selector).data(function (d) { return [d]; });
                    tooltipRect.enter().append("path").classed(PulseChart.TooltipRect.class, true);
                    tooltipRect
                        .attr("display", function (d) { return d.popupInfo ? "inherit" : "none"; })
                        .style('fill', this.data.settings.popup.color)
                        .attr('d', function (d) {
                        var path = [
                            {
                                "x": -2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * marginTop) : 0,
                            },
                            {
                                "x": -2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop + height)) : height,
                            },
                            {
                                "x": width - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop + height)) : height,
                            },
                            {
                                "x": width - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * marginTop) : 0,
                            }
                        ];
                        return line(path);
                    });
                    var tooltipTriangle = tooltipRoot.selectAll(PulseChart.TooltipTriangle.selector).data(function (d) { return [d]; });
                    tooltipTriangle.enter().append("path").classed(PulseChart.TooltipTriangle.class, true);
                    tooltipTriangle
                        .style('fill', this.data.settings.popup.color)
                        .attr('d', function (d) {
                        var path = [
                            {
                                "x": width / 2 - 5 - d.popupInfo.offsetX,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * marginTop) : 0,
                            },
                            {
                                "x": width / 2 - d.popupInfo.offsetX,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop - 5)) : -5,
                            },
                            {
                                "x": width / 2 + 5 - d.popupInfo.offsetX,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * marginTop) : 0,
                            },
                        ];
                        return line(path);
                    })
                        .style('stroke-width', "1px");
                    var tooltipLine = tooltipRoot.selectAll(PulseChart.TooltipLine.selector).data(function (d) { return [d]; });
                    tooltipLine.enter().append("path").classed(PulseChart.TooltipLine.class, true);
                    tooltipLine
                        .style('fill', this.data.settings.popup.color)
                        .style('stroke', this.data.settings.popup.color)
                        .style('stroke-width', "1px")
                        .attr('d', function (d) {
                        var path = [
                            {
                                "x": width / 2 - d.popupInfo.offsetX,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ?
                                    yScales[d.groupIndex](d.y) + tooltipShiftY(d.y, d.groupIndex) - d.eventSize :
                                    yScales[d.groupIndex](d.y) - tooltipShiftY(d.y, d.groupIndex) + d.eventSize,
                            },
                            {
                                "x": width / 2 - d.popupInfo.offsetX,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * marginTop) : 0,
                            }];
                        return line(path);
                    });
                    var isShowTime = this.data.settings.popup.showTime;
                    var isShowTitle = this.data.settings.popup.showTitle;
                    var timeRect = tooltipRoot.selectAll(PulseChart.TooltipTimeRect.selector).data(function (d) { return [d]; });
                    var timeDisplayStyle = { "display": isShowTime ? undefined : "none" };
                    timeRect.enter().append("path").classed(PulseChart.TooltipTimeRect.class, true);
                    timeRect
                        .style("fill", this.data.settings.popup.timeFill)
                        .style(timeDisplayStyle)
                        .attr('d', function (d) {
                        var path = [
                            {
                                "x": width - _this.data.widthOfTooltipValueLabel - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop + height)) : 0,
                            },
                            {
                                "x": width - _this.data.widthOfTooltipValueLabel - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex)
                                    ? (-1 * (marginTop + height - PulseChart.DefaultTooltipSettings.timeHeight))
                                    : PulseChart.DefaultTooltipSettings.timeHeight,
                            },
                            {
                                "x": width - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex)
                                    ? (-1 * (marginTop + height - PulseChart.DefaultTooltipSettings.timeHeight))
                                    : PulseChart.DefaultTooltipSettings.timeHeight,
                            },
                            {
                                "x": width - 2,
                                "y": _this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop + height)) : 0,
                            }
                        ];
                        return line(path);
                    });
                    var time = tooltipRoot.selectAll(PulseChart.TooltipTime.selector).data(function (d) { return [d]; });
                    time.enter().append("text").classed(PulseChart.TooltipTime.class, true);
                    time
                        .style(PulseChart.ConvertTextPropertiesToStyle(PulseChart.GetPopupValueTextProperties()))
                        .style(timeDisplayStyle)
                        .style("fill", this.data.settings.popup.timeColor)
                        .attr("x", function (d) { return width - _this.data.widthOfTooltipValueLabel; })
                        .attr("y", function (d) { return _this.isHigherMiddle(d.y, d.groupIndex)
                        ? (-1 * (marginTop + height - PulseChart.DefaultTooltipSettings.timeHeight + 3))
                        : PulseChart.DefaultTooltipSettings.timeHeight - 3; })
                        .text(function (d) { return d.popupInfo.value; });
                    var titleDisplayStyle = { "display": isShowTitle ? undefined : "none" };
                    var title = tooltipRoot.selectAll(PulseChart.TooltipTitle.selector).data(function (d) { return [d]; });
                    title.enter().append("text").classed(PulseChart.TooltipTitle.class, true);
                    title
                        .style(titleDisplayStyle)
                        .style(PulseChart.ConvertTextPropertiesToStyle(PulseChart.GetPopupTitleTextProperties()))
                        .style("fill", this.data.settings.popup.fontColor)
                        .attr("x", function (d) { return PulseChart.PopupTextPadding; })
                        .attr("y", function (d) {
                        return (_this.isHigherMiddle(d.y, d.groupIndex) ? (-1 * (marginTop + height - 12)) : 12) + PulseChart.PopupTextPadding;
                    })
                        .text(function (d) {
                        if (!d.popupInfo) {
                            return "";
                        }
                        var maxWidth = width - PulseChart.PopupTextPadding * 2 -
                            (isShowTime ? (_this.data.widthOfTooltipValueLabel - PulseChart.PopupTextPadding) : 0) - 10;
                        return TextMeasurementService.getTailoredTextOrDefault(PulseChart.GetPopupTitleTextProperties(d.popupInfo.title), maxWidth);
                    });
                    var getDescriptionDimenstions = function (d) {
                        var shiftY = PulseChart.PopupTextPadding + _this.data.settings.popup.fontSize;
                        var descriptionYOffset = shiftY + PulseChart.DefaultTooltipSettings.timeHeight;
                        if (d.popupInfo) {
                            shiftY = ((isShowTitle && d.popupInfo.title) || (isShowTime && d.popupInfo.value)) ? descriptionYOffset : shiftY;
                        }
                        return {
                            y: _this.isHigherMiddle(d.y, d.groupIndex)
                                ? (-1 * (marginTop + height - shiftY))
                                : shiftY,
                            x: PulseChart.PopupTextPadding,
                            width: width - PulseChart.PopupTextPadding * 2,
                            height: height - shiftY,
                        };
                    };
                    var description = tooltipRoot.selectAll(PulseChart.TooltipDescription.selector).data(function (d) { return [d]; });
                    description.enter().append("text").classed(PulseChart.TooltipDescription.class, true);
                    description
                        .style(PulseChart.ConvertTextPropertiesToStyle(PulseChart.GetPopupDescriptionTextProperties(null, this.data.settings.popup.fontSize)))
                        .style("fill", this.data.settings.popup.fontColor)
                        .text(function (d) { return d.popupInfo && d.popupInfo.description; })
                        .attr("y", function (d) {
                        var descriptionDimenstions = getDescriptionDimenstions(d);
                        var el = d3.select(this)[0][0];
                        TextMeasurementService.wordBreak(el, descriptionDimenstions.width, descriptionDimenstions.height);
                        return 0;
                    })
                        .attr("transform", function (d) {
                        var descriptionDimenstions = getDescriptionDimenstions(d);
                        return visuals.SVGUtil.translate(0, descriptionDimenstions.y);
                    });
                    description.selectAll("tspan").attr("x", PulseChart.PopupTextPadding);
                    tooltipRoot
                        .exit()
                        .remove();
                };
                PulseChart.prototype.isHigherMiddle = function (value, groupIndex) {
                    if (this.data.settings.popup.alwaysOnTop) {
                        return true;
                    }
                    if (this.data.yScales.length > 1) {
                        return groupIndex === 0;
                    }
                    var domain = this.data.commonYScale.domain(), minValue = d3.min(domain), middleValue = Math.abs((d3.max(domain) - minValue) / 2);
                    middleValue = middleValue === 0
                        ? middleValue
                        : minValue + middleValue;
                    return value >= middleValue;
                };
                PulseChart.getObjectsFromDataView = function (dataView) {
                    if (!dataView ||
                        !dataView.metadata ||
                        !dataView.metadata.columns ||
                        !dataView.metadata.objects) {
                        return null;
                    }
                    return dataView.metadata.objects;
                };
                PulseChart.parseSettings = function (dataView, colors, columns) {
                    var settings = {}, objects = PulseChart.getObjectsFromDataView(dataView);
                    settings.xAxis = this.getAxisXSettings(objects, colors);
                    settings.yAxis = this.getAxisYSettings(objects, colors);
                    settings.popup = this.getPopupSettings(objects, colors);
                    settings.dots = this.getDotsSettings(objects, colors);
                    settings.series = this.getSeriesSettings(objects, colors);
                    settings.gaps = this.getGapsSettings(objects);
                    settings.playback = this.getPlaybackSettings(objects, colors);
                    settings.runnerCounter = this.getRunnerCounterSettings(objects, colors, columns);
                    return settings;
                };
                PulseChart.getPopupSettings = function (objects, colors) {
                    var show = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["show"], PulseChart.DefaultSettings.popup.show);
                    var alwaysOnTop = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["alwaysOnTop"], PulseChart.DefaultSettings.popup.alwaysOnTop);
                    var width = Math.max(PulseChart.PopupMinWidth, Math.min(PulseChart.PopupMaxWidth, powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["width"], PulseChart.DefaultSettings.popup.width)));
                    var height = Math.max(PulseChart.PopupMinHeight, Math.min(PulseChart.PopupMaxHeight, powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["height"], PulseChart.DefaultSettings.popup.height)));
                    var colorHelper = new visuals.ColorHelper(colors, PulseChart.Properties["popup"]["color"], PulseChart.DefaultSettings.popup.color);
                    var color = colorHelper.getColorForMeasure(objects, "");
                    var fontSize = parseInt(powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["fontSize"], PulseChart.DefaultSettings.popup.fontSize), 10);
                    var fontColorHelper = new visuals.ColorHelper(colors, PulseChart.Properties["popup"]["fontColor"], PulseChart.DefaultSettings.popup.fontColor);
                    var fontColor = fontColorHelper.getColorForMeasure(objects, "");
                    var showTime = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["showTime"], PulseChart.DefaultSettings.popup.showTime);
                    var showTitle = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["popup"]["showTitle"], PulseChart.DefaultSettings.popup.showTitle);
                    var timeColorHelper = new visuals.ColorHelper(colors, PulseChart.Properties["popup"]["timeColor"], PulseChart.DefaultSettings.popup.timeColor);
                    var timeColor = timeColorHelper.getColorForMeasure(objects, "");
                    var timeFillHelper = new visuals.ColorHelper(colors, PulseChart.Properties["popup"]["timeFill"], PulseChart.DefaultSettings.popup.timeFill);
                    var timeFill = timeFillHelper.getColorForMeasure(objects, "");
                    return {
                        show: show,
                        alwaysOnTop: alwaysOnTop,
                        width: width,
                        height: height,
                        color: color,
                        fontSize: fontSize,
                        fontColor: fontColor,
                        showTime: showTime,
                        showTitle: showTitle,
                        timeColor: timeColor,
                        timeFill: timeFill,
                    };
                };
                PulseChart.getDotsSettings = function (objects, colors) {
                    var properties = PulseChart.Properties["dots"], defaultSettings = PulseChart.DefaultSettings.dots;
                    var colorHelper = new visuals.ColorHelper(colors, properties["color"], defaultSettings.color);
                    var color = colorHelper.getColorForMeasure(objects, "");
                    var minSize = Math.max(0, Math.min(9999, powerbi.DataViewObjects.getValue(objects, properties["minSize"], defaultSettings.minSize)));
                    var maxSize = Math.max(minSize, Math.min(9999, powerbi.DataViewObjects.getValue(objects, properties["maxSize"], defaultSettings.maxSize)));
                    var size = Math.max(minSize, Math.min(maxSize, powerbi.DataViewObjects.getValue(objects, properties["size"], defaultSettings.size)));
                    var transparency = Math.max(0, Math.min(100, powerbi.DataViewObjects.getValue(objects, properties["transparency"], defaultSettings.transparency)));
                    return {
                        color: color,
                        size: size,
                        minSize: minSize,
                        maxSize: maxSize,
                        transparency: transparency,
                    };
                };
                PulseChart.getSeriesSettings = function (objects, colors) {
                    var width = Math.max(1, Math.min(100, powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["series"]["width"], PulseChart.DefaultSettings.series.width)));
                    var colorHelper = new visuals.ColorHelper(colors, PulseChart.Properties["series"]["fill"], PulseChart.DefaultSettings.series.fill);
                    var fill = colorHelper.getColorForMeasure(objects, "");
                    /*var showByDefault = DataViewObjects.getValue<boolean>(
                        objects,
                        PulseChart.Properties["series"]["showByDefault"],
                        PulseChart.DefaultSettings.series.showByDefault);*/
                    return {
                        width: width,
                        fill: fill,
                    };
                };
                PulseChart.getGapsSettings = function (objects) {
                    var show = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["gaps"]["show"], PulseChart.DefaultSettings.gaps.show);
                    var visibleGapsPercentage = Math.max(1, Math.min(100, powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["gaps"]["transparency"], PulseChart.DefaultSettings.gaps.visibleGapsPercentage)));
                    return {
                        show: show,
                        visibleGapsPercentage: visibleGapsPercentage
                    };
                };
                PulseChart.getAxisXSettings = function (objects, colors) {
                    var properties = PulseChart.Properties["xAxis"], defaultSettings = PulseChart.DefaultSettings.xAxis;
                    var color = new visuals.ColorHelper(colors, properties["color"], defaultSettings.color).getColorForMeasure(objects, "");
                    var fontColor = new visuals.ColorHelper(colors, properties["fontColor"], defaultSettings.fontColor).getColorForMeasure(objects, "");
                    var show = powerbi.DataViewObjects.getValue(objects, properties["show"], defaultSettings.show);
                    var position = powerbi.DataViewObjects.getValue(objects, properties["position"], defaultSettings.position);
                    var backgroundColor = new visuals.ColorHelper(colors, properties["backgroundColor"], defaultSettings.backgroundColor).getColorForMeasure(objects, "");
                    return {
                        show: show,
                        position: position,
                        color: color,
                        fontColor: fontColor,
                        backgroundColor: backgroundColor
                    };
                };
                PulseChart.getAxisYSettings = function (objects, colors) {
                    var properties = PulseChart.Properties["yAxis"], defaultSettings = PulseChart.DefaultSettings.yAxis;
                    var colorHelper = new visuals.ColorHelper(colors, properties["color"], defaultSettings.color);
                    var color = colorHelper.getColorForMeasure(objects, "");
                    var fontColorHelper = new visuals.ColorHelper(colors, properties["fontColor"], defaultSettings.fontColor);
                    var fontColor = fontColorHelper.getColorForMeasure(objects, "");
                    var show = powerbi.DataViewObjects.getValue(objects, properties["show"], defaultSettings.show);
                    return {
                        color: color,
                        fontColor: fontColor,
                        show: show,
                    };
                };
                PulseChart.getPlaybackSettings = function (objects, colors) {
                    var playbackSettings = {};
                    var properties = PulseChart.Properties["playback"], defaultSettings = PulseChart.DefaultSettings.playback;
                    playbackSettings.autoplay = powerbi.DataViewObjects.getValue(objects, properties["autoplay"], defaultSettings.autoplay);
                    playbackSettings.playSpeed = Math.max(1, Math.min(99999, powerbi.DataViewObjects.getValue(objects, properties["playSpeed"], defaultSettings.playSpeed)));
                    playbackSettings.pauseDuration = Math.max(0, Math.min(9999, powerbi.DataViewObjects.getValue(objects, properties["pauseDuration"], defaultSettings.pauseDuration)));
                    playbackSettings.autoplayPauseDuration = Math.max(1, Math.min(9999, powerbi.DataViewObjects.getValue(objects, properties["autoplayPauseDuration"], defaultSettings.autoplayPauseDuration)));
                    var colorHelper = new visuals.ColorHelper(colors, properties["color"], defaultSettings.color);
                    playbackSettings.color = colorHelper.getColorForMeasure(objects, "");
                    var position = powerbi.DataViewObjects.getValue(objects, properties["position"], "");
                    if (position.length > 3) {
                        try {
                            playbackSettings.position = JSON.parse(position);
                        }
                        catch (ex) { }
                    }
                    playbackSettings.position = playbackSettings.position || defaultSettings.position;
                    return playbackSettings;
                };
                PulseChart.getRunnerCounterSettings = function (objects, colors, columns) {
                    var show = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["runnerCounter"]["show"], PulseChart.DefaultSettings.runnerCounter.show);
                    var label = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["runnerCounter"]["label"], columns.RunnerCounter && columns.RunnerCounter.source && columns.RunnerCounter.source.displayName
                        || PulseChart.DefaultSettings.runnerCounter.label);
                    var position = powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["runnerCounter"]["position"], PulseChart.DefaultSettings.runnerCounter.position);
                    var fontSize = parseInt(powerbi.DataViewObjects.getValue(objects, PulseChart.Properties["runnerCounter"]["fontSize"], PulseChart.DefaultSettings.runnerCounter.fontSize), 10);
                    var fontColor = new visuals.ColorHelper(colors, PulseChart.Properties["runnerCounter"]["fontColor"], PulseChart.DefaultSettings.runnerCounter.fontColor)
                        .getColorForMeasure(objects, "");
                    return {
                        show: show,
                        label: label,
                        position: position,
                        fontSize: fontSize,
                        fontColor: fontColor
                    };
                };
                PulseChart.prototype.clearAll = function (hide) {
                    this.gaps.selectAll(PulseChart.Gap.selector).remove();
                    if (this.animationHandler) {
                        this.animationHandler.reset();
                        this.animationHandler.clear();
                    }
                    if (hide) {
                        this.svg.style('display', "none");
                    }
                    this.clearChart();
                };
                PulseChart.prototype.clearChart = function () {
                    this.clearSelection();
                    this.hideAnimationDot();
                    this.chart.selectAll(PulseChart.Line.selector).remove();
                    this.chart.selectAll(PulseChart.Dot.selector).remove();
                };
                PulseChart.prototype.clearRedundant = function (position) {
                    var _this = this;
                    if (!this.data) {
                        return;
                    }
                    var popups = this.chart.selectAll(PulseChart.Tooltip.selector).filter(function (data) {
                        return data.seriesIndex < position.series || data.seriesIndex === position.series
                            && _this.data.series[data.seriesIndex].data[position.index].value >= data.value;
                    });
                    var selectedPopupsIds = popups.data().map(function (data) { return data.identity; });
                    this.clearSelection();
                    if (selectedPopupsIds.length) {
                        var popupsSelecting = selectedPopupsIds.map(function (id) { return _this.selectionManager.select(id, true); });
                        $.when.apply(null, popupsSelecting)
                            .then(function (selectionIds) { return _this.setSelection(selectionIds); });
                    }
                    this.chart.selectAll(PulseChart.Line.selector).remove();
                    this.chart.selectAll(PulseChart.Dot.selector).remove();
                };
                PulseChart.prototype.enumerateObjectInstances = function (options) {
                    var enumeration = new visuals.ObjectEnumerationBuilder();
                    switch (options.objectName) {
                        case "general": {
                            this.readGeneralInstance(enumeration);
                            break;
                        }
                        case "popup": {
                            this.readPopupInstance(enumeration);
                            break;
                        }
                        case "dots": {
                            this.readDotsInstance(enumeration);
                            break;
                        }
                        case "xAxis": {
                            this.xAxisInstance(enumeration);
                            break;
                        }
                        case "yAxis": {
                            this.yAxisInstance(enumeration);
                            break;
                        }
                        case "series": {
                            this.readSeriesInstance(enumeration);
                            break;
                        }
                        case "gaps": {
                            this.readGapsInstance(enumeration);
                            break;
                        }
                        case "playback": {
                            this.readPlaybackInstance(enumeration);
                            break;
                        }
                        case "runnerCounter": {
                            this.readRunnerCounterInstance(enumeration);
                            break;
                        }
                    }
                    return enumeration.complete();
                };
                PulseChart.prototype.getSettings = function (name) {
                    if (this.data && this.data.settings && this.data.settings[name]) {
                        return this.data.settings[name];
                    }
                    return PulseChart.DefaultSettings[name];
                };
                PulseChart.prototype.readGeneralInstance = function (enumeration) {
                    var instance = {
                        objectName: "general",
                        displayName: "general",
                        selector: null,
                        properties: {}
                    };
                    enumeration.pushInstance(instance);
                };
                PulseChart.prototype.readPopupInstance = function (enumeration) {
                    var settings = this.getSettings("popup");
                    var popup = {
                        objectName: "popup",
                        displayName: "popup",
                        selector: null,
                        properties: {
                            show: settings.show,
                            alwaysOnTop: settings.alwaysOnTop,
                            width: settings.width,
                            height: settings.height,
                            color: settings.color,
                            fontColor: settings.fontColor,
                            fontSize: settings.fontSize,
                            showTime: settings.showTime,
                            showTitle: settings.showTitle,
                            timeColor: settings.timeColor,
                            timeFill: settings.timeFill,
                        }
                    };
                    enumeration.pushInstance(popup);
                };
                PulseChart.prototype.readDotsInstance = function (enumeration) {
                    var settings = this.getSettings("dots");
                    var instance = {
                        objectName: "dots",
                        displayName: "Dots",
                        selector: null,
                        properties: {
                            color: settings.color,
                            size: settings.size,
                            minSize: settings.minSize,
                            maxSize: settings.maxSize,
                            transparency: settings.transparency,
                        }
                    };
                    enumeration.pushInstance(instance);
                };
                PulseChart.prototype.xAxisInstance = function (enumeration) {
                    var settings = this.getSettings("xAxis");
                    enumeration.pushInstance({
                        objectName: "xAxis",
                        displayName: "xAxis",
                        selector: null,
                        properties: {
                            show: settings.show,
                            position: settings.position,
                            color: settings.color,
                            fontColor: settings.fontColor,
                            backgroundColor: settings.backgroundColor
                        }
                    });
                };
                PulseChart.prototype.yAxisInstance = function (enumeration) {
                    var settings = this.getSettings("yAxis");
                    enumeration.pushInstance({
                        objectName: "yAxis",
                        displayName: "yAxis",
                        selector: null,
                        properties: {
                            color: settings.color,
                            fontColor: settings.fontColor,
                            show: settings.show
                        }
                    });
                };
                PulseChart.prototype.readSeriesInstance = function (enumeration) {
                    var settings = this.getSettings("series");
                    var series = {
                        objectName: "series",
                        displayName: "series",
                        selector: null,
                        properties: {
                            fill: settings.fill,
                            width: settings.width,
                        }
                    };
                    enumeration.pushInstance(series);
                };
                PulseChart.prototype.readGapsInstance = function (enumeration) {
                    var settings = this.getSettings("gaps");
                    var gaps = {
                        objectName: "gaps",
                        selector: null,
                        properties: {
                            show: settings.show,
                            transparency: settings.visibleGapsPercentage //visibleGapsPercentage
                        }
                    };
                    enumeration.pushInstance(gaps);
                };
                PulseChart.prototype.readPlaybackInstance = function (enumeration) {
                    var settings = this.getSettings("playback");
                    enumeration.pushInstance({
                        objectName: "playback",
                        displayName: "playback",
                        selector: null,
                        properties: {
                            autoplay: settings.autoplay,
                            playSpeed: settings.playSpeed,
                            pauseDuration: settings.pauseDuration,
                            autoplayPauseDuration: settings.autoplayPauseDuration,
                            color: settings.color,
                        }
                    });
                };
                PulseChart.prototype.readRunnerCounterInstance = function (enumeration) {
                    var runnerCounterSettings = this.getSettings("runnerCounter");
                    var instance = {
                        objectName: "runnerCounter",
                        selector: null,
                        properties: {}
                    };
                    if (this.data &&
                        this.data.columns &&
                        this.data.columns.RunnerCounter) {
                        instance.properties = {
                            show: runnerCounterSettings.show,
                            label: runnerCounterSettings.label,
                            position: runnerCounterSettings.position,
                            fontSize: runnerCounterSettings.fontSize,
                            fontColor: runnerCounterSettings.fontColor
                        };
                    }
                    enumeration.pushInstance(instance);
                };
                PulseChart.prototype.destroy = function () {
                    this.data = null;
                    this.clearAll(true);
                };
                PulseChart.RoleDisplayNames = {
                    Timestamp: "Timestamp",
                    Category: "Category",
                    Value: "Value",
                    EventTitle: "Event Title",
                    EventDescription: "Event Description",
                    EventSize: "Event Size",
                    RunnerCounter: "Runner Counter",
                };
                PulseChart.RoleNames = _.mapValues(PulseChart.RoleDisplayNames, function (x, i) { return i; });
                PulseChart.capabilities = {
                    dataRoles: [
                        {
                            displayName: PulseChart.RoleDisplayNames.Timestamp,
                            name: PulseChart.RoleNames.Timestamp,
                            kind: powerbi.VisualDataRoleKind.Grouping
                        },
                        {
                            displayName: PulseChart.RoleDisplayNames.Value,
                            name: PulseChart.RoleNames.Value,
                            kind: powerbi.VisualDataRoleKind.Measure
                        },
                        /* Temporary disabled
                        {
                            displayName: PulseChart.RoleDisplayNames.Category,
                            name: PulseChart.RoleNames.Category,
                            kind: powerbi.VisualDataRoleKind.Grouping
                        },*/
                        {
                            displayName: PulseChart.RoleDisplayNames.EventTitle,
                            name: PulseChart.RoleNames.EventTitle,
                            kind: powerbi.VisualDataRoleKind.GroupingOrMeasure
                        },
                        {
                            displayName: PulseChart.RoleDisplayNames.EventDescription,
                            name: PulseChart.RoleNames.EventDescription,
                            kind: powerbi.VisualDataRoleKind.GroupingOrMeasure
                        },
                        {
                            displayName: PulseChart.RoleDisplayNames.EventSize,
                            name: PulseChart.RoleNames.EventSize,
                            kind: powerbi.VisualDataRoleKind.GroupingOrMeasure
                        },
                        {
                            displayName: PulseChart.RoleDisplayNames.RunnerCounter,
                            name: PulseChart.RoleNames.RunnerCounter,
                            kind: powerbi.VisualDataRoleKind.GroupingOrMeasure
                        },
                    ],
                    dataViewMappings: [{
                            conditions: [
                                {
                                    Timestamp: { max: 1 },
                                    Value: { max: 1 },
                                    Category: { max: 1 },
                                    EventTitle: { max: 1 },
                                    EventDescription: { max: 1 },
                                    EventSize: { max: 1 },
                                    RunnerCounter: { max: 1 },
                                }
                            ],
                            categorical: {
                                categories: {
                                    for: { in: PulseChart.RoleNames.Timestamp },
                                    dataReductionAlgorithm: { top: { count: 10000 } }
                                },
                                values: {
                                    group: {
                                        by: PulseChart.RoleNames.Category,
                                        select: [
                                            { bind: { to: PulseChart.RoleNames.Value } },
                                            { bind: { to: PulseChart.RoleNames.EventTitle } },
                                            { bind: { to: PulseChart.RoleNames.EventDescription } },
                                            { bind: { to: PulseChart.RoleNames.EventSize } },
                                            { bind: { to: PulseChart.RoleNames.RunnerCounter } },
                                        ],
                                        dataReductionAlgorithm: { top: {} }
                                    },
                                },
                            },
                        }],
                    objects: {
                        series: {
                            displayName: "Series",
                            description: "Series",
                            properties: {
                                fill: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_Fill'),
                                    type: {
                                        fill: {
                                            solid: {
                                                color: true
                                            }
                                        }
                                    }
                                },
                                width: {
                                    displayName: 'Width',
                                    type: {
                                        numeric: true
                                    }
                                },
                            }
                        },
                        gaps: {
                            displayName: "Gaps",
                            description: "Gaps",
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                transparency: {
                                    displayName: 'Visible gaps',
                                    type: { numeric: true }
                                },
                            }
                        },
                        general: {
                            displayName: 'General',
                            properties: {
                                formatString: { type: { formatting: { formatString: true } } },
                                fill: {
                                    displayName: 'Background color',
                                    type: { fill: { solid: { color: true } } }
                                }
                            }
                        },
                        popup: {
                            displayName: 'Popup',
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                alwaysOnTop: {
                                    displayName: 'Always on top',
                                    type: { bool: true }
                                },
                                width: {
                                    displayName: 'Width',
                                    type: {
                                        numeric: true
                                    }
                                },
                                height: {
                                    displayName: 'Height',
                                    type: {
                                        numeric: true
                                    }
                                },
                                color: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_Fill'),
                                    type: { fill: { solid: { color: true } } }
                                },
                                fontSize: {
                                    displayName: "Text size",
                                    type: { formatting: { fontSize: true } }
                                },
                                fontColor: {
                                    displayName: "Text color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                showTime: {
                                    displayName: 'Show time',
                                    type: { bool: true }
                                },
                                showTitle: {
                                    displayName: 'Show title',
                                    type: { bool: true }
                                },
                                timeColor: {
                                    displayName: "Time color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                timeFill: {
                                    displayName: "Time fill",
                                    type: { fill: { solid: { color: true } } }
                                },
                            }
                        },
                        dots: {
                            displayName: 'Dots',
                            properties: {
                                color: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_Fill'),
                                    type: { fill: { solid: { color: true } } }
                                },
                                minSize: {
                                    displayName: "Min Size",
                                    type: { numeric: true }
                                },
                                maxSize: {
                                    displayName: "Max Size",
                                    type: { numeric: true }
                                },
                                size: {
                                    displayName: "Default Size",
                                    type: { numeric: true }
                                },
                                transparency: {
                                    displayName: 'Transparency',
                                    type: { numeric: true }
                                },
                            }
                        },
                        xAxis: {
                            displayName: powerbi.data.createDisplayNameGetter('Visual_XAxis'),
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                position: {
                                    displayName: "Position",
                                    type: { enumeration: createEnumTypeFromEnum(XAxisPosition) }
                                },
                                fontColor: {
                                    displayName: "Font Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                color: {
                                    displayName: "Axis Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                backgroundColor: {
                                    displayName: "Background Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                            }
                        },
                        yAxis: {
                            displayName: powerbi.data.createDisplayNameGetter('Visual_YAxis'),
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                fontColor: {
                                    displayName: "Font Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                color: {
                                    displayName: "Axis Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                            }
                        },
                        playback: {
                            displayName: 'Playback',
                            properties: {
                                autoplay: {
                                    displayName: "Autoplay",
                                    type: { bool: true }
                                },
                                playSpeed: {
                                    displayName: "Speed (dots/sec)",
                                    type: { numeric: true }
                                },
                                pauseDuration: {
                                    displayName: "Pause Duration",
                                    type: { numeric: true }
                                },
                                autoplayPauseDuration: {
                                    displayName: "Start Delay",
                                    type: { numeric: true }
                                },
                                color: {
                                    displayName: "Buttons Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                                position: {
                                    displayName: "Position",
                                    type: { text: true }
                                }
                            }
                        },
                        runnerCounter: {
                            displayName: 'Runner Counter',
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                label: {
                                    displayName: "Label",
                                    type: { text: true }
                                },
                                position: {
                                    displayName: "Position",
                                    type: { enumeration: createEnumTypeFromEnum(RunnerCounterPosition) }
                                },
                                fontSize: {
                                    displayName: "Text Size",
                                    type: { formatting: { fontSize: true } }
                                },
                                fontColor: {
                                    displayName: "Font Color",
                                    type: { fill: { solid: { color: true } } }
                                },
                            }
                        },
                    },
                    sorting: {
                        implicit: {
                            clauses: [{
                                    role: PulseChart.RoleNames.Timestamp,
                                    direction: 1 //SortDirection.Ascending
                                }]
                        }
                    },
                    supportsHighlight: true
                };
                PulseChart.Properties = PulseChart.getProperties(PulseChart.capabilities);
                PulseChart.DefaultMargin = {
                    top: 20,
                    bottom: 20,
                    right: 25,
                    left: 25,
                };
                PulseChart.DefaultViewport = {
                    width: 50,
                    height: 50
                };
                PulseChart.PlaybackButtonsHeight = 26;
                PulseChart.PopupMinHeight = 20;
                PulseChart.PopupMinWidth = 20;
                PulseChart.PopupMaxHeight = 200;
                PulseChart.PopupMaxWidth = 2000;
                PulseChart.MaxWidthOfYAxis = 50;
                PulseChart.PopupTextPadding = 3;
                PulseChart.XAxisTickSpace = 15;
                PulseChart.XAxisTickHeight = 16;
                PulseChart.MinimumTicksToRotate = 3;
                PulseChart.AxisTickRotateAngle = -35;
                PulseChart.DefaultSettings = {
                    precision: 0,
                    popup: {
                        show: true,
                        alwaysOnTop: false,
                        width: 100,
                        height: 80,
                        color: "#808181",
                        fontSize: 10,
                        fontColor: 'white',
                        showTime: true,
                        showTitle: true,
                        timeColor: 'white',
                        timeFill: '#010101',
                    },
                    dots: {
                        color: "#808181",
                        size: 5,
                        minSize: 5,
                        maxSize: 20,
                        transparency: 25,
                    },
                    gaps: {
                        show: false,
                        visibleGapsPercentage: 1
                    },
                    series: {
                        fill: '#3779B7',
                        width: 2,
                    },
                    xAxis: {
                        color: "#777777",
                        fontColor: "#777777",
                        position: XAxisPosition.Center,
                        show: true,
                        dateFormat: PulseChartXAxisDateFormat.TimeOnly,
                        backgroundColor: "#E1F2F7"
                    },
                    yAxis: {
                        color: "#777777",
                        fontColor: "#777777",
                        show: true
                    },
                    playback: {
                        autoplay: false,
                        playSpeed: 5,
                        pauseDuration: 10,
                        autoplayPauseDuration: 0,
                        color: "#777",
                        position: null,
                    },
                    runnerCounter: {
                        show: true,
                        label: "",
                        position: RunnerCounterPosition.TopRight,
                        fontSize: 13,
                        fontColor: "#777777"
                    },
                    formatStringProperty: PulseChart.Properties["general"]["formatString"]
                };
                PulseChart.DefaultTooltipSettings = {
                    dataPointColor: "#808181",
                    marginTop: 20,
                    timeHeight: 15,
                };
                PulseChart.MaxGapCount = 100;
                PulseChart.MinGapWidth = _.object([[
                        PulseChartXAxisDateFormat.DateOnly, 60 * 1000 * 24], [
                        PulseChartXAxisDateFormat.TimeOnly, 60 * 1000],
                ], undefined);
                PulseChart.Chart = createClassAndSelector('chart');
                PulseChart.Line = createClassAndSelector('line');
                PulseChart.LineContainer = createClassAndSelector('lineContainer');
                PulseChart.LineNode = createClassAndSelector('lineNode');
                PulseChart.XAxisNode = createClassAndSelector('xAxisNode');
                PulseChart.Dot = createClassAndSelector('dot');
                PulseChart.DotsContainer = createClassAndSelector('dotsContainer');
                PulseChart.Tooltip = createClassAndSelector('Tooltip');
                PulseChart.TooltipRect = createClassAndSelector('tooltipRect');
                PulseChart.TooltipTriangle = createClassAndSelector('tooltipTriangle');
                PulseChart.Gaps = createClassAndSelector("gaps");
                PulseChart.Gap = createClassAndSelector("gap");
                PulseChart.GapNode = createClassAndSelector("gapNode");
                PulseChart.TooltipLine = createClassAndSelector('tooltipLine');
                PulseChart.TooltipTime = createClassAndSelector('tooltipTime');
                PulseChart.TooltipTimeRect = createClassAndSelector('tooltipTimeRect');
                PulseChart.TooltipTitle = createClassAndSelector('tooltipTitle');
                PulseChart.TooltipDescription = createClassAndSelector('tooltipDescription');
                PulseChart.TooltipContainer = createClassAndSelector('tooltipContainer');
                PulseChart.AnimationDot = createClassAndSelector('animationDot');
                return PulseChart;
            }());
            PulseChart1459209850231.PulseChart = PulseChart;
            var PulseAnimatorStates;
            (function (PulseAnimatorStates) {
                PulseAnimatorStates[PulseAnimatorStates["Ready"] = 0] = "Ready";
                PulseAnimatorStates[PulseAnimatorStates["Play"] = 1] = "Play";
                PulseAnimatorStates[PulseAnimatorStates["Paused"] = 2] = "Paused";
                PulseAnimatorStates[PulseAnimatorStates["Stopped"] = 3] = "Stopped";
            })(PulseAnimatorStates || (PulseAnimatorStates = {}));
            var PulseAnimator = (function () {
                function PulseAnimator(chart, svg) {
                    var _this = this;
                    this.animationPlayingIndex = 0;
                    this.runnerCounterTopLeftPosition = 180;
                    this.isAutoPlayed = false;
                    this.isPositionWasSaved = false;
                    this.chart = chart;
                    this.svg = svg;
                    this.setDefaultValues();
                    var container = this.container = this.svg.append('g');
                    this.animationPlay = container.append('g').classed(PulseAnimator.AnimationPlay.class, true);
                    this.animationPlay
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationPlay
                        .call(PulseChart.AddOnTouchClick, function () { return _this.play(); });
                    this.animationPlay
                        .append("path")
                        .attr("d", "M12 2c5.514 0 10 4.486 10 10s-4.486 10-10 10-10-4.486-10-10 4.486-10 10-10zm0-2c-6.627 0-12 5.373-12 12s5.373 12 12 12 12-5.373 12-12-5.373-12-12-12zm-3 17v-10l9 5.146-9 4.854z");
                    this.animationPause = container.append('g').classed(PulseAnimator.AnimationPause.class, true);
                    this.animationPause
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationPause
                        .call(PulseChart.AddOnTouchClick, function () { return _this.stop(); });
                    this.animationPause
                        .append("path")
                        .attr("d", "M12 2c5.514 0 10 4.486 10 10s-4.486 10-10 10-10-4.486-10-10 4.486-10 10-10zm0-2c-6.627 0-12 5.373-12 12s5.373 12 12 12 12-5.373 12-12-5.373-12-12-12zm-1 17h-3v-10h3v10zm5-10h-3v10h3v-10z");
                    this.animationReset = container.append('g').classed(PulseAnimator.AnimationReset.class, true);
                    this.animationReset
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationReset
                        .call(PulseChart.AddOnTouchClick, function () { return _this.reset(); });
                    this.animationReset
                        .append("path")
                        .attr("d", "M22 12c0 5.514-4.486 10-10 10s-10-4.486-10-10 4.486-10 10-10 10 4.486 10 10zm-22 0c0 6.627 5.373 12 12 12s12-5.373 12-12-5.373-12-12-12-12 5.373-12 12zm13 0l5-4v8l-5-4zm-5 0l5-4v8l-5-4zm-2 4h2v-8h-2v8z");
                    /* Prev */
                    this.animationPrev = container.append('g').classed(PulseAnimator.AnimationPrev.class, true);
                    this.animationPrev
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationPrev
                        .call(PulseChart.AddOnTouchClick, function () { return _this.prev(); });
                    this.animationPrev
                        .append("path")
                        .attr("d", "M9.5 12l7.5-4.5v9l-7.5-4.5zm-4.5 0l6.5 4v-1.634l-3.943-2.366 3.943-2.366v-1.634l-6.5 4zm17 0c0 5.514-4.486 10-10 10s-10-4.486-10-10 4.486-10 10-10 10 4.486 10 10zm-22 0c0 6.627 5.373 12 12 12s12-5.373 12-12-5.373-12-12-12-12 5.373-12 12z");
                    /* Next */
                    this.animationNext = container.append('g').classed(PulseAnimator.AnimationNext.class, true);
                    this.animationNext
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationNext
                        .call(PulseChart.AddOnTouchClick, function () { return _this.next(); });
                    this.animationNext
                        .append("path")
                        .attr("d", "M7 16.5v-9l7.5 4.5-7.5 4.5zm5.5-8.5v1.634l3.943 2.366-3.943 2.366v1.634l6.5-4-6.5-4zm-.5-6c5.514 0 10 4.486 10 10s-4.486 10-10 10-10-4.486-10-10 4.486-10 10-10zm0-2c-6.627 0-12 5.373-12 12s5.373 12 12 12 12-5.373 12-12-5.373-12-12-12z")
                        .attr("rotate", 180);
                    /* ToEnd */
                    this.animationToEnd = container.append('g').classed(PulseAnimator.AnimationToEnd.class, true);
                    this.animationToEnd
                        .append("circle")
                        .attr("cx", 12)
                        .attr("cy", 12)
                        .attr("r", 10)
                        .attr("fill", "transparent");
                    this.animationToEnd
                        .call(PulseChart.AddOnTouchClick, function () { return _this.toEnd(); });
                    this.animationToEnd
                        .append("path")
                        .attr("d", "M12 2c5.514 0 10 4.486 10 10s-4.486 10-10 10-10-4.486-10-10 4.486-10 10-10zm0-2c-6.627 0-12 5.373-12 12s5.373 12 12 12 12-5.373 12-12-5.373-12-12-12zm-6 16v-8l5 4-5 4zm5 0v-8l5 4-5 4zm7-8h-2v8h2v-8z");
                    this.runnerCounter = container.append('g').classed(PulseAnimator.RunnerCounter.class, true);
                    this.runnerCounterText = this.runnerCounter.append('text');
                    this.setControlsColor(PulseAnimator.DefaultControlsColor);
                }
                Object.defineProperty(PulseAnimator, "AnimationMinPosition", {
                    get: function () {
                        return { series: 0, index: 0 };
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "runnerCounterPosition", {
                    get: function () {
                        return this.chart.data.settings.runnerCounter.position;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "maxTextWidthOfRunnerCounterValue", {
                    get: function () {
                        var top = this.runnerCounterPosition === RunnerCounterPosition.TopLeft || this.runnerCounterPosition === RunnerCounterPosition.TopRight;
                        return this.chart.viewport.width - (top ? this.runnerCounterTopLeftPosition : 0);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "isAnimated", {
                    get: function () {
                        return (this.animatorState === PulseAnimatorStates.Paused) ||
                            (this.animatorState === PulseAnimatorStates.Play) ||
                            (this.animatorState === PulseAnimatorStates.Stopped);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "isPlaying", {
                    get: function () {
                        return this.animatorState === PulseAnimatorStates.Play;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "isPaused", {
                    get: function () {
                        return this.animatorState === PulseAnimatorStates.Paused;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "isStopped", {
                    get: function () {
                        return this.animatorState === PulseAnimatorStates.Stopped;
                    },
                    enumerable: true,
                    configurable: true
                });
                PulseAnimator.prototype.setDefaultValues = function () {
                    this.position = PulseAnimator.AnimationMinPosition;
                    this.animatorState = PulseAnimatorStates.Ready;
                    this.runnerCounterValue = null;
                };
                PulseAnimator.prototype.render = function () {
                    this.renderControls();
                    this.disableControls();
                    if (!this.chart.isAutoPlay) {
                        this.isAutoPlayed = true;
                        if (this.savedPosition) {
                            this.savedPosition = null;
                        }
                    }
                    if (this.chart.isAutoPlay && this.isAutoPlayed
                        && this.animatorState === PulseAnimatorStates.Play
                        && !this.isPositionWasSaved
                        && !_.isEqual(this.autoPlayPosition, this.savedPosition)) {
                        this.chart.stopAnimation();
                        this.isAutoPlayed = false;
                        this.isPositionWasSaved = true;
                        this.animatorState = PulseAnimatorStates.Ready;
                    }
                    if (this.animatorState === PulseAnimatorStates.Play) {
                        this.play();
                    }
                    else if (this.chart.isAutoPlay && !this.isAutoPlayed && (this.animatorState === PulseAnimatorStates.Ready)) {
                        //console.log("loaded " + JSON.stringify(this.savedPosition));
                        this.autoPlayPosition = this.savedPosition;
                        this.isAutoPlayed = true;
                        if (this.savedPosition
                            && this.savedPosition.series < this.chart.data.series.length
                            && this.savedPosition.index < this.chart.data.series[this.savedPosition.series].data.length) {
                            this.position = this.savedPosition;
                        }
                        this.play(this.chart.autoplayPauseDuration);
                    }
                    else {
                        this.chart.renderChart();
                    }
                };
                PulseAnimator.prototype.setControlsColor = function (color) {
                    this.color = color;
                };
                PulseAnimator.prototype.renderControls = function () {
                    this.show();
                    this.animationPlay
                        .attr('transform', visuals.SVGUtil.translate(0, 0))
                        .attr("fill", this.color);
                    this.animationPause
                        .attr('transform', visuals.SVGUtil.translate(30, 0))
                        .attr("fill", this.color);
                    this.animationReset
                        .attr('transform', visuals.SVGUtil.translate(60, 0))
                        .attr("fill", this.color);
                    this.animationPrev
                        .attr('transform', visuals.SVGUtil.translate(90, 0))
                        .attr("fill", this.color);
                    this.animationNext
                        .attr('transform', visuals.SVGUtil.translate(120, 0))
                        .attr("fill", this.color);
                    this.animationToEnd
                        .attr('transform', visuals.SVGUtil.translate(150, 0))
                        .attr("fill", this.color);
                    this.runnerCounter
                        .attr('fill', this.color)
                        .attr('transform', visuals.SVGUtil.translate(this.runnerCounterPosition === RunnerCounterPosition.TopLeft
                        ? this.runnerCounterTopLeftPosition
                        : this.chart.viewport.width - 2, this.chart.data.runnerCounterHeight / 2 + 7));
                    this.runnerCounterText
                        .style('text-anchor', this.runnerCounterPosition === RunnerCounterPosition.TopLeft ? "start" : "end");
                    if (this.chart.data && this.chart.data.settings) {
                        this.runnerCounterText.style(PulseChart.ConvertTextPropertiesToStyle(PulseChart.GetRunnerCounterTextProperties(null, this.chart.data.settings.runnerCounter.fontSize)));
                        this.runnerCounterText.style('fill', this.chart.data.settings.runnerCounter.fontColor);
                    }
                    this.drawCounterValue();
                };
                PulseAnimator.setControlVisiblity = function (element, isVisible, isDisabled) {
                    if (isDisabled === void 0) { isDisabled = false; }
                    element
                        .style('opacity', isVisible ? PulseAnimator.DefaultOpacity : PulseAnimator.DimmedOpacity);
                    if (isVisible) {
                        element.attr('display', "inline");
                    }
                    else if (isDisabled) {
                        element.attr('display', "none");
                    }
                };
                PulseAnimator.prototype.disableControls = function () {
                    var showRunner = this.chart.data && this.chart.data.settings && this.chart.data.settings.runnerCounter.show;
                    PulseAnimator.setControlVisiblity(this.animationReset, true);
                    PulseAnimator.setControlVisiblity(this.animationToEnd, true);
                    switch (this.animatorState) {
                        case PulseAnimatorStates.Play:
                            PulseAnimator.setControlVisiblity(this.animationPlay, false);
                            PulseAnimator.setControlVisiblity(this.animationPrev, true);
                            PulseAnimator.setControlVisiblity(this.animationNext, true);
                            PulseAnimator.setControlVisiblity(this.animationPause, true);
                            PulseAnimator.setControlVisiblity(this.runnerCounter, showRunner, true);
                            break;
                        case PulseAnimatorStates.Paused:
                            PulseAnimator.setControlVisiblity(this.animationPlay, true);
                            PulseAnimator.setControlVisiblity(this.animationPause, true);
                            PulseAnimator.setControlVisiblity(this.animationPrev, true);
                            PulseAnimator.setControlVisiblity(this.animationNext, true);
                            PulseAnimator.setControlVisiblity(this.runnerCounter, showRunner, true);
                            break;
                        case PulseAnimatorStates.Stopped:
                            PulseAnimator.setControlVisiblity(this.animationPlay, true);
                            PulseAnimator.setControlVisiblity(this.animationPrev, true);
                            PulseAnimator.setControlVisiblity(this.animationNext, true);
                            PulseAnimator.setControlVisiblity(this.runnerCounter, showRunner, true);
                            PulseAnimator.setControlVisiblity(this.animationPause, false);
                            break;
                        case PulseAnimatorStates.Ready:
                            PulseAnimator.setControlVisiblity(this.animationPlay, true);
                            PulseAnimator.setControlVisiblity(this.animationPrev, false);
                            PulseAnimator.setControlVisiblity(this.animationNext, false);
                            PulseAnimator.setControlVisiblity(this.animationPause, false);
                            PulseAnimator.setControlVisiblity(this.runnerCounter, false, true);
                            break;
                        default:
                            PulseAnimator.setControlVisiblity(this.animationPlay, true);
                            PulseAnimator.setControlVisiblity(this.animationPrev, false);
                            PulseAnimator.setControlVisiblity(this.animationNext, false);
                            PulseAnimator.setControlVisiblity(this.animationPause, false);
                            PulseAnimator.setControlVisiblity(this.runnerCounter, false, true);
                            break;
                    }
                };
                PulseAnimator.prototype.show = function () {
                    this.container.style('display', 'inline');
                };
                PulseAnimator.prototype.setRunnerCounterValue = function (index) {
                    var dataPoint = this.chart.data
                        && this.chart.data.series
                        && this.chart.data.series[this.position.series]
                        && this.chart.data.series[this.position.series].data
                        && this.chart.data.series[this.position.series].data[$.isNumeric(index) ? index : this.flooredPosition.index];
                    var runnerCounterValue = (dataPoint && dataPoint.runnerCounterValue != null)
                        ? dataPoint.runnerCounterValue
                        : "";
                    if (dataPoint && dataPoint.runnerCounterFormatString) {
                        var runnerCounterformatter = visuals.valueFormatter.create({ format: dataPoint.runnerCounterFormatString });
                        runnerCounterValue = runnerCounterformatter.format(runnerCounterValue);
                    }
                    this.runnerCounterValue = this.chart.data.settings.runnerCounter.label + " " + runnerCounterValue;
                    this.drawCounterValue();
                };
                PulseAnimator.prototype.drawCounterValue = function () {
                    var progressText = "" + this.runnerCounterValue;
                    this.runnerCounterText.text(progressText);
                    TextMeasurementService.svgEllipsis(this.runnerCounterText.node(), this.maxTextWidthOfRunnerCounterValue);
                };
                PulseAnimator.prototype.play = function (delay, renderDuringPlaying) {
                    if (delay === void 0) { delay = 0; }
                    if (renderDuringPlaying === void 0) { renderDuringPlaying = false; }
                    if (this.animatorState === PulseAnimatorStates.Play && !renderDuringPlaying) {
                        return;
                    }
                    if (this.animatorState === PulseAnimatorStates.Ready) {
                        this.animationPlayingIndex++;
                        this.chart.clearChart();
                    }
                    if (this.chart.isAnimationIndexLast(this.position)) {
                        this.playNext();
                        return;
                    }
                    if (this.animatorState === PulseAnimatorStates.Paused) {
                        this.chart.clearSelection();
                    }
                    this.animatorState = PulseAnimatorStates.Play;
                    this.chart.renderChart();
                    this.chart.playAnimation(delay);
                    this.disableControls();
                };
                PulseAnimator.prototype.playNext = function () {
                    this.pause();
                    if (this.chart.isAnimationSeriesLast(this.position)) {
                        this.setDefaultValues();
                        this.chart.clearSelection();
                    }
                    else {
                        this.position = {
                            series: this.position.series + 1,
                            index: PulseAnimator.AnimationMinPosition.index,
                        };
                        this.play();
                    }
                };
                PulseAnimator.prototype.pause = function () {
                    if (this.animatorState === PulseAnimatorStates.Play) {
                        this.animatorState = PulseAnimatorStates.Paused;
                        this.chart.pauseAnimation();
                    }
                    this.disableControls();
                };
                PulseAnimator.prototype.reset = function () {
                    this.clearTimeouts();
                    this.chart.stopAnimation();
                    this.chart.clearSelection();
                    this.chart.clearChart();
                    this.setDefaultValues();
                    this.animatorState = PulseAnimatorStates.Stopped;
                    this.disableControls();
                    this.savedPosition = null;
                };
                PulseAnimator.prototype.next = function () {
                    if (!this.isAnimated) {
                        return;
                    }
                    this.stop();
                    var newPosition = this.chart.findNextPoint(this.position);
                    if (newPosition) {
                        this.position = newPosition;
                        this.chart.renderChart();
                    }
                    else {
                        this.toEnd();
                    }
                };
                PulseAnimator.prototype.prev = function () {
                    if (!this.isAnimated) {
                        return;
                    }
                    this.stop();
                    var newPosition = this.chart.findPrevPoint(this.position);
                    if (newPosition) {
                        this.chart.clearRedundant(newPosition);
                        this.position = newPosition;
                        this.chart.renderChart();
                    }
                    else {
                        this.reset();
                    }
                };
                PulseAnimator.prototype.toEnd = function () {
                    this.savedPosition = null;
                    this.chart.stopAnimation();
                    this.chart.clearSelection();
                    this.chart.clearChart();
                    this.setDefaultValues();
                    this.disableControls();
                    this.chart.renderChart();
                };
                PulseAnimator.prototype.stop = function () {
                    if (!this.isAnimated) {
                        return;
                    }
                    this.drawCounterValue();
                    this.savedPosition = this.position;
                    this.chart.stopAnimation();
                    this.animatorState = PulseAnimatorStates.Stopped;
                    this.disableControls();
                };
                Object.defineProperty(PulseAnimator.prototype, "position", {
                    get: function () {
                        return this.positionValue;
                    },
                    set: function (position) {
                        this.positionValue = position;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "flooredPosition", {
                    get: function () {
                        return this.position && { series: this.position.series, index: Math.floor(this.position.index) };
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(PulseAnimator.prototype, "savedPosition", {
                    get: function () {
                        return this.chart.data
                            && this.chart.data.settings
                            && this.chart.data.settings.playback
                            && this.chart.data.settings.playback.position;
                    },
                    set: function (position) {
                        if (!this.chart.isAutoPlay) {
                            position = null;
                        }
                        if (this.chart.data && this.chart.data.settings && this.chart.data.settings.playback) {
                            //console.log("saved " + JSON.stringify(position));
                            this.isPositionWasSaved = true;
                            if (this.chart.data && this.chart.data.settings && this.chart.data.settings.playback) {
                                this.chart.data.settings.playback.position = position;
                            }
                            this.chart.host.persistProperties({
                                merge: [{
                                        objectName: "playback",
                                        selector: null,
                                        properties: { position: position && JSON.stringify(position) || "" }
                                    }] });
                        }
                    },
                    enumerable: true,
                    configurable: true
                });
                PulseAnimator.prototype.clear = function () {
                    if (this.isAnimated) {
                        this.chart.stopAnimation();
                    }
                    this.setDefaultValues();
                    this.container.style('display', 'none');
                };
                PulseAnimator.prototype.clearTimeouts = function () {
                    clearTimeout(this.chart.handleSelectionTimeout);
                };
                PulseAnimator.AnimationPlay = createClassAndSelector('animationPlay');
                PulseAnimator.AnimationPause = createClassAndSelector('animationPause');
                PulseAnimator.AnimationReset = createClassAndSelector('animationReset');
                PulseAnimator.AnimationToEnd = createClassAndSelector('animationToEnd');
                PulseAnimator.AnimationPrev = createClassAndSelector('animationPrev');
                PulseAnimator.AnimationNext = createClassAndSelector('animationNext');
                PulseAnimator.RunnerCounter = createClassAndSelector('runnerCounter');
                //private static ControlsDuration = 250;
                PulseAnimator.DimmedOpacity = 0.25;
                PulseAnimator.DefaultOpacity = 1;
                PulseAnimator.DefaultControlsColor = "#777";
                return PulseAnimator;
            }());
            PulseChart1459209850231.PulseAnimator = PulseAnimator;
            var PulseChartDataLabelUtils;
            (function (PulseChartDataLabelUtils) {
                function getDefaultPulseChartLabelSettings() {
                    return {
                        show: false,
                        position: 0 /* Above */,
                        displayUnits: 0,
                        precision: defaultLabelPrecision,
                        labelColor: defaultLabelColor,
                        fontSize: DefaultFontSizeInPt,
                        labelDensity: defaultLabelDensity,
                    };
                }
                PulseChartDataLabelUtils.getDefaultPulseChartLabelSettings = getDefaultPulseChartLabelSettings;
            })(PulseChartDataLabelUtils = PulseChart1459209850231.PulseChartDataLabelUtils || (PulseChart1459209850231.PulseChartDataLabelUtils = {}));
            var PulseChartAxisPropertiesHelper;
            (function (PulseChartAxisPropertiesHelper) {
                function getCategoryAxisProperties(dataViewMetadata, axisTitleOnByDefault) {
                    var toReturn = {};
                    if (!dataViewMetadata) {
                        return toReturn;
                    }
                    var objects = dataViewMetadata.objects;
                    if (objects) {
                        var categoryAxisObject = objects['categoryAxis'];
                        if (categoryAxisObject) {
                            toReturn = {
                                show: categoryAxisObject['show'],
                                axisType: categoryAxisObject['axisType'],
                                axisScale: categoryAxisObject['axisScale'],
                                start: categoryAxisObject['start'],
                                end: categoryAxisObject['end'],
                                showAxisTitle: categoryAxisObject['showAxisTitle'] == null ? axisTitleOnByDefault : categoryAxisObject['showAxisTitle'],
                                axisStyle: categoryAxisObject['axisStyle'],
                                labelColor: categoryAxisObject['labelColor'],
                                labelDisplayUnits: categoryAxisObject['labelDisplayUnits'],
                                labelPrecision: categoryAxisObject['labelPrecision'],
                                duration: categoryAxisObject['duration'],
                            };
                        }
                    }
                    return toReturn;
                }
                PulseChartAxisPropertiesHelper.getCategoryAxisProperties = getCategoryAxisProperties;
                function getValueAxisProperties(dataViewMetadata, axisTitleOnByDefault) {
                    var toReturn = {};
                    if (!dataViewMetadata) {
                        return toReturn;
                    }
                    var objects = dataViewMetadata.objects;
                    if (objects) {
                        var valueAxisObject = objects['valueAxis'];
                        if (valueAxisObject) {
                            toReturn = {
                                show: valueAxisObject['show'],
                                position: valueAxisObject['position'],
                                axisScale: valueAxisObject['axisScale'],
                                start: valueAxisObject['start'],
                                end: valueAxisObject['end'],
                                showAxisTitle: valueAxisObject['showAxisTitle'] == null ? axisTitleOnByDefault : valueAxisObject['showAxisTitle'],
                                axisStyle: valueAxisObject['axisStyle'],
                                labelColor: valueAxisObject['labelColor'],
                                labelDisplayUnits: valueAxisObject['labelDisplayUnits'],
                                labelPrecision: valueAxisObject['labelPrecision'],
                                secShow: valueAxisObject['secShow'],
                                secPosition: valueAxisObject['secPosition'],
                                secAxisScale: valueAxisObject['secAxisScale'],
                                secStart: valueAxisObject['secStart'],
                                secEnd: valueAxisObject['secEnd'],
                                secShowAxisTitle: valueAxisObject['secShowAxisTitle'],
                                secAxisStyle: valueAxisObject['secAxisStyle'],
                                secLabelColor: valueAxisObject['secLabelColor'],
                                secLabelDisplayUnits: valueAxisObject['secLabelDisplayUnits'],
                                secLabelPrecision: valueAxisObject['secLabelPrecision'],
                            };
                        }
                    }
                    return toReturn;
                }
                PulseChartAxisPropertiesHelper.getValueAxisProperties = getValueAxisProperties;
            })(PulseChartAxisPropertiesHelper = PulseChart1459209850231.PulseChartAxisPropertiesHelper || (PulseChart1459209850231.PulseChartAxisPropertiesHelper = {}));
        })(PulseChart1459209850231 = visuals.PulseChart1459209850231 || (visuals.PulseChart1459209850231 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.PulseChart1459209850231 = {
                name: 'PulseChart1459209850231',
                class: 'PulseChart1459209850231',
                capabilities: powerbi.visuals.PulseChart1459209850231.PulseChart.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.PulseChart1459209850231.PulseChart(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
